# -*- coding: utf-8 -*-
"""Created on Tue Oct 18 10:35:51 2022.

@author: Barney

"""

# import pytest
import unittest
from unittest import TestCase

from wsimod.core import constants
from wsimod.nodes.wtw import WTW, WWTW, FWTW


class MyTestClass(TestCase):
    def setUp(self):
        """"""
        constants.set_simple_pollutants()

    def assertDictAlmostEqual(self, d1, d2, accuracy=19):
        """

        Args:
            d1:
            d2:
            accuracy:
        """
        for d in [d1, d2]:
            for key, item in d.items():
                d[key] = round(item, accuracy)
        self.assertDictEqual(d1, d2)

    def test_excess(self):
        wtw = WTW(
            name="",
            treatment_throughput_capacity=10,
        )
        wtw.current_input["volume"] = 8
        self.assertEqual(2, wtw.get_excess_throughput())

    def test_treat(self):
        constants.set_simple_pollutants()
        wtw = WTW(
            name="",
            treatment_throughput_capacity=10,
        )
        wtw.current_input = {
            "volume": 8,
            "phosphate": 5,
            "temperature": constants.DECAY_REFERENCE_TEMPERATURE,
        }
        wtw.treat_current_input()
        self.assertEqual(
            8 * wtw.process_parameters["volume"]["constant"], wtw.treated["volume"]
        )
        self.assertEqual(
            5 * wtw.process_parameters["phosphate"]["constant"],
            wtw.treated["phosphate"],
        )
        self.assertEqual(8 * wtw.liquor_multiplier["volume"], wtw.liquor["volume"])
        self.assertEqual(
            5 * wtw.liquor_multiplier["phosphate"], wtw.liquor["phosphate"]
        )
        self.assertEqual(
            5
            - 5 * wtw.process_parameters["phosphate"]["constant"]
            - 5 * wtw.liquor_multiplier["phosphate"],
            wtw.solids["phosphate"],
        )

    def test_override(self):
        wtw = WTW(
            name="",
            treatment_throughput_capacity=10,
            percent_solids=0.1,
            liquor_multiplier={"volume": 0.05, "phosphate": 0.5},
            process_parameters={"phosphate": {"constant": 0.1, "exponent": 1.001}},
        )

        wtw.apply_overrides({"percent_solids": 0.05})
        self.assertAlmostEqual(wtw.process_parameters["volume"]["constant"], 0.9)
        self.assertEqual(wtw.percent_solids, 0.05)

        wtw.apply_overrides(
            {"percent_solids": 0.1, "liquor_multiplier": {"volume": 0.1}}
        )

        self.assertEqual(wtw.process_parameters["volume"]["constant"], 0.8)
        self.assertEqual(wtw.liquor_multiplier["volume"], 0.1)
        self.assertEqual(wtw.liquor_multiplier["phosphate"], 0.5)

        wtw.apply_overrides(
            {
                "percent_solids": 0.1,
                "liquor_multiplier": {"volume": 0.1, "phosphate": 0.01},
            }
        )
        self.assertEqual(wtw.liquor_multiplier["phosphate"], 0.01)

        wtw.apply_overrides({"process_parameters": {"phosphate": {"constant": 0.01}}})
        self.assertEqual(wtw.process_parameters["phosphate"]["constant"], 0.01)
        self.assertEqual(wtw.process_parameters["phosphate"]["exponent"], 1.001)

        overrides = {
            "process_parameters": {"phosphate": {"exponent": 1.01}},
            "liquor_multiplier": {"phosphate": 0.1},
            "percent_solids": 0.1,
            "treatment_throughput_capacity": 20,
            "name": "new_name",
        }
        wtw.apply_overrides(overrides)
        self.assertSetEqual(set(overrides.keys()), set(["name"]))
        self.assertEqual(wtw.treatment_throughput_capacity, 20)

    def test_wwtw_overrides(self):
        wwtw = WWTW(name="")
        vol = wwtw.process_parameters["volume"]["constant"]
        wwtw.apply_overrides(
            {
                "treatment_throughput_capacity": 20,
                "process_parameters": {"phosphate": {"constant": 0.01}},
                "stormwater_storage_capacity": 100,
            }
        )
        self.assertEqual(wwtw.treatment_throughput_capacity, 20)
        self.assertEqual(wwtw.process_parameters["phosphate"]["constant"], 0.01)
        self.assertEqual(wwtw.process_parameters["volume"]["constant"], vol)
        self.assertEqual(wwtw.stormwater_storage_capacity, 100)

    def test_fwtw_overrides(self):
        fwtw = FWTW(name="")
        vol = fwtw.process_parameters["volume"]["constant"]
        fwtw.apply_overrides(
            {
                "treatment_throughput_capacity": 20,
                "process_parameters": {"phosphate": {"constant": 0.02}},
                "service_reservoir_storage_capacity": 100,
                "service_reservoir_storage_area": 34.7,
                "service_reservoir_storage_elevation": 68.2,
            }
        )
        self.assertEqual(fwtw.treatment_throughput_capacity, 20)
        self.assertEqual(fwtw.process_parameters["phosphate"]["constant"], 0.02)
        self.assertEqual(fwtw.process_parameters["volume"]["constant"], vol)
        self.assertEqual(fwtw.service_reservoir_storage_capacity, 100)
        self.assertEqual(fwtw.service_reservoir_tank.capacity, 100)
        self.assertEqual(fwtw.service_reservoir_storage_area, 34.7)
        self.assertEqual(fwtw.service_reservoir_tank.area, 34.7)
        self.assertEqual(fwtw.service_reservoir_storage_elevation, 68.2)
        self.assertEqual(fwtw.service_reservoir_tank.datum, 68.2)


if __name__ == "__main__":
    unittest.main()



from wsimod.extensions import register_node_patch


@register_node_patch("my_dist", "push_check_handler", item="Node")
def custom_handler_function(self, vqip, *args, **kwargs):
    """A custom `push_check_handler` function.

    Call the default handler for the "Node" item.
    """
    print("I reached a custom handler")
    return self.push_check_handler["default"](vqip)



# -*- coding: utf-8 -*-
"""Created on Wed Apr  7 08:43:32 2021.

@author: Barney

Converted to totals on Thur Apr 21 2022
"""
import logging
from typing import Any, Dict

from wsimod.core import constants
from wsimod.core.core import WSIObj


class Node(WSIObj):
    """"""

    def __init_subclass__(cls, **kwargs):
        """Adds all subclasses to the nodes registry."""
        super().__init_subclass__(**kwargs)
        if cls.__name__ in NODES_REGISTRY:
            logging.warning(f"Overwriting {cls.__name__} in NODES_REGISTRY with {cls}")

        NODES_REGISTRY[cls.__name__] = cls

    def __init__(self, name, data_input_dict=None):
        """Base class for CWSD nodes. Constructs all the necessary attributes for the
        node object.

        Args:
            name (str): Name of node
            data_input_dict (dict, optional): Dictionary of data inputs relevant for
                the node. Keys are tuples where first value is the name of the
                variable to read from the dict and the second value is the time.
                Defaults to None.

        Examples:
            >>> my_node = nodes.Node(name = 'london_river_junction')

        Key assumptions:
            - No physical processes represented, can be used as a junction.

        Input data and parameter requirements:
            - All nodes require a `name`
        """
        node_types = list(NODES_REGISTRY.keys())

        # Default essential parameters
        # Dictionary of arcs
        self.in_arcs = {}
        self.out_arcs = {}
        self.in_arcs_type = {x: {} for x in node_types}
        self.out_arcs_type = {x: {} for x in node_types}

        # Set parameters
        self.name = name
        self.t = None
        self.data_input_dict = data_input_dict

        # Initiailise default handlers
        self.pull_set_handler = {"default": self.pull_distributed}
        self.push_set_handler = {
            "default": lambda x: self.push_distributed(
                x, of_type=["Node", "River", "Waste", "Reservoir"]
            )
        }
        self.pull_check_handler = {"default": self.pull_check_basic}
        self.push_check_handler = {
            "default": lambda x: self.push_check_basic(
                x, of_type=["Node", "River", "Waste", "Reservoir"]
            )
        }
        super().__init__()

        # Mass balance checking
        self.mass_balance_in = [self.total_in]
        self.mass_balance_out = [self.total_out]
        self.mass_balance_ds = [lambda: self.empty_vqip()]

    def apply_overrides(self, overrides: Dict[str, Any] = {}) -> None:
        """Apply overrides to the node.

        The Node does not have any overwriteable parameters. So if any
        overrides are passed up to the node, this means that there are unused
        parameters from the Node subclass, which is flagged.

        Args:
            overrides (dict, optional): Dictionary of overrides. Defaults to {}.
        """
        # overrides data_input_dict
        from wsimod.orchestration.model import read_csv

        content = overrides.pop("data_input_dict", self.data_input_dict)
        if isinstance(content, str):
            self.data_input_dict = read_csv(content)
        elif not content:
            pass
        else:
            raise RuntimeError("Not recognised format for data_input_dict")

        if len(overrides) > 0:
            print(f"No override behaviour defined for: {overrides.keys()}")

    def total_in(self):
        """Sum flow and pollutant amounts entering a node via in_arcs.

        Returns:
            in_ (dict): Summed VQIP of in_arcs

        Examples:
            >>> node_inflow = my_node.total_in()
        """
        in_ = self.empty_vqip()
        for arc in self.in_arcs.values():
            in_ = self.sum_vqip(in_, arc.vqip_out)

        return in_

    def total_out(self):
        """Sum flow and pollutant amounts leaving a node via out_arcs.

        Returns:
            out_ (dict): Summed VQIP of out_arcs

        Examples:
            >>> node_outflow = my_node.total_out()
        """
        out_ = self.empty_vqip()
        for arc in self.out_arcs.values():
            out_ = self.sum_vqip(out_, arc.vqip_in)

        return out_

    def node_mass_balance(self):
        """Wrapper for core.py/WSIObj/mass_balance. Tracks change in mass balance.

        Returns:
            in_ (dict): A VQIP of the total from mass_balance_in functions
            ds_ (dict): A VQIP of the total from mass_balance_ds functions
            out_ (dict): A VQIP of the total from mass_balance_out functions

        Examples:
            >>> node_in, node_out, node_ds = my_node.node_mass_balance()
        """
        in_, ds_, out_ = self.mass_balance()
        return in_, ds_, out_

    def pull_set(self, vqip, tag="default"):
        """Receives pull set requests from arcs and passes request to query handler.

        Args:
            vqip (dict): the VQIP pull request (by default, only the 'volume' key is
                needed).
            tag (str, optional): optional message to direct query_handler which pull
                function to call. Defaults to 'default'.

        Returns:
            (dict): VQIP received from query_handler

        Examples:
            >>> water_received = my_node.pull_set({'volume' : 10})
        """
        return self.query_handler(self.pull_set_handler, vqip, tag)

    def push_set(self, vqip, tag="default"):
        """Receives push set requests from arcs and passes request to query handler.

        Args:
            vqip (_type_): the VQIP push request
            tag (str, optional): optional message to direct query_handler which pull
                function to call. Defaults to 'default'.

        Returns:
            (dict): VQIP not received from query_handler

        Examples:
            water_not_pushed = my_node.push_set(wastewater_vqip)
        """
        return self.query_handler(self.push_set_handler, vqip, tag)

    def pull_check(self, vqip=None, tag="default"):
        """Receives pull check requests from arcs and passes request to query handler.

        Args:
            vqip (dict, optional): the VQIP pull check (by default, only the
                'volume' key is used). Defaults to None, which returns all available
                water to pull.
            tag (str, optional): optional message to direct query_handler which pull
                function to call. Defaults to 'default'.

        Returns:
            (dict): VQIP available from query_handler

        Examples:
            >>> water_available = my_node.pull_check({'volume' : 10})
            >>> total_water_available = my_node.pull_check()
        """
        return self.query_handler(self.pull_check_handler, vqip, tag)

    def push_check(self, vqip=None, tag="default"):
        """Receives push check requests from arcs and passes request to query handler.

        Args:
            vqip (dict, optional): the VQIP push check. Defaults to None, which
                returns all available capacity to push
            tag (str, optional): optional message to direct query_handler which pull
                function to call. Defaults to 'default'

        Returns:
            (dict): VQIP available to push from query_handler

        Examples:
            >>> total_available_push_capacity = my_node.push_check()
            >>> available_push_capacity = my_node.push_check(wastewater_vqip)
        """
        return self.query_handler(self.push_check_handler, vqip, tag)

    def get_direction_arcs(self, direction, of_type=None):
        """Identify arcs to/from all attached nodes in a given direction.

        Args:
            direction (str): can be either 'pull' or 'push' to send checks to
                receiving or contributing nodes
            of_type (str or list) : optional, can be specified to send checks only
                to nodes of a given type (must be a subclass in nodes.py)

        Returns:
            f (str): Either 'send_pull_check' or 'send_push_check' depending on
                direction
            arcs (list): List of arc objects

        Raises:
            Message if no direction is specified

        Examples:
            >>> arcs_to_push_to = my_node.get_direction_arcs('push')
            >>> arcs_to_pull_from = my_node.get_direction_arcs('pull')
            >>> arcs_from_reservoirs = my_node.get_direction_arcs('pull', of_type =
                'Reservoir')
        """
        if of_type is None:
            # Return all arcs
            if direction == "pull":
                arcs = list(self.in_arcs.values())
                f = "send_pull_check"
            elif direction == "push":
                arcs = list(self.out_arcs.values())
                f = "send_push_check"
            else:
                print("No direction")

        else:
            if isinstance(of_type, str):
                of_type = [of_type]

            # Assign arcs/function based on parameters
            arcs = []
            if direction == "pull":
                for type_ in of_type:
                    arcs += list(self.in_arcs_type[type_].values())
                f = "send_pull_check"
            elif direction == "push":
                for type_ in of_type:
                    arcs += list(self.out_arcs_type[type_].values())
                f = "send_push_check"
            else:
                print("No direction")

        return f, arcs

    def get_connected(self, direction="pull", of_type=None, tag="default"):
        """Send push/pull checks to all attached arcs in a given direction.

        Args:
            direction (str, optional): The type of check to send to all attached
                nodes. Can be 'push' or 'pull'. The default is 'pull'.
            of_type (str or list) : optional, can be specified to send checks only
                to nodes of a given type (must be a subclass in nodes.py)
            tag (str, optional): optional message to direct query_handler which pull
                function to call. Defaults to 'default'.

        Returns:
            connected (dict) :
                Dictionary containing keys:
                'avail': (float) - total available volume for push/pull
                'priority': (float) - total (availability * preference)
                                    of attached arcs
                'allocation': (dict) - contains all attached arcs in specified
                                direction and respective (availability * preference)

        Examples:
            >>> vqip_available_to_pull = my_node.get_direction_arcs()
            >>> vqip_available_to_push = my_node.get_direction_arcs('push')
            >>> avail_reservoir_vqip = my_node.get_direction_arcs('pull',
                                                          of_type = 'Reservoir')
            >>> avail_sewer_push_to_sewers = my_node.get_direction_arcs('push',
                                                                of_type = 'Sewer',
                                                                tag = 'Sewer')
        """
        # Initialise connected dict
        connected = {"avail": 0, "priority": 0, "allocation": {}, "capacity": {}}

        # Get arcs
        f, arcs = self.get_direction_arcs(direction, of_type)

        # Iterate over arcs, updating connected dict
        for arc in arcs:
            avail = getattr(arc, f)(tag=tag)["volume"]
            if avail < constants.FLOAT_ACCURACY:
                avail = 0  # Improves convergence
            connected["avail"] += avail
            preference = arc.preference
            connected["priority"] += avail * preference
            connected["allocation"][arc.name] = avail * preference
            connected["capacity"][arc.name] = avail

        return connected

    def query_handler(self, handler, ip, tag):
        """Sends all push/pull requests/checks using the handler (i.e., ensures the
        correct function is used that lines up with 'tag').

        Args:
            handler (dict): contains all push/pull requests for various tags
            ip (vqip): the vqip request
            tag (str): describes what type of push/pull request should be called

        Returns:
            (dict): the VQIP reply from push/pull request

        Raises:
            Message if no functions are defined for tag and if request/check
            function fails
        """
        try:
            return handler[tag](ip)
        except Exception:
            if tag not in handler.keys():
                print("No functions defined for " + tag)
                return handler[tag](ip)
            else:
                print("Some other error")
                return handler[tag](ip)

    def pull_distributed(self, vqip, of_type=None, tag="default"):
        """Send pull requests to all (or specified by type) nodes connecting to self.
        Iterate until request is met or maximum iterations are hit. Streamlines if only
        one in_arc exists.

        Args:
            vqip (dict): Total amount to pull (by default, only the
                'volume' key is used)
            of_type (str or list) : optional, can be specified to send checks only
                to nodes of a given type (must be a subclass in nodes.py)
            tag (str, optional): optional message to direct query_handler which pull
                function to call. Defaults to 'default'.

        Returns:
            pulled (dict): VQIP of combined pulled water
        """
        if len(self.in_arcs) == 1:
            # If only one in_arc, just pull from that
            if of_type is None:
                pulled = next(iter(self.in_arcs.values())).send_pull_request(
                    vqip, tag=tag
                )
            elif any(
                [x in of_type for x, y in self.in_arcs_type.items() if len(y) > 0]
            ):
                pulled = next(iter(self.in_arcs.values())).send_pull_request(
                    vqip, tag=tag
                )
            else:
                # No viable out arcs
                pulled = self.empty_vqip()
        else:
            # Pull in proportion from connected by priority

            # Initialise pulled, deficit, connected, iter_
            pulled = self.empty_vqip()
            deficit = vqip["volume"]
            connected = self.get_connected(direction="pull", of_type=of_type, tag=tag)
            iter_ = 0

            # Iterate over sending nodes until deficit met
            while (
                (deficit > constants.FLOAT_ACCURACY)
                & (connected["avail"] > constants.FLOAT_ACCURACY)
            ) & (iter_ < constants.MAXITER):
                # Pull from connected
                for key, allocation in connected["allocation"].items():
                    received = self.in_arcs[key].send_pull_request(
                        {"volume": deficit * allocation / connected["priority"]},
                        tag=tag,
                    )
                    pulled = self.sum_vqip(pulled, received)

                # Update deficit, connected and iter_
                deficit = vqip["volume"] - pulled["volume"]
                connected = self.get_connected(
                    direction="pull", of_type=of_type, tag=tag
                )
                iter_ += 1

            if iter_ == constants.MAXITER:
                print("Maxiter reached in {0} at {1}".format(self.name, self.t))
        return pulled

    def push_distributed(self, vqip, of_type=None, tag="default"):
        """Send push requests to all (or specified by type) nodes connecting to self.
        Iterate until request is met or maximum iterations are hit. Streamlines if only
        one in_arc exists.

        Args:
            vqip (dict): Total amount to push
            of_type (str or list) : optional, can be specified to send checks only
                to nodes of a given type (must be a subclass in nodes.py)
            tag (str, optional): optional message to direct query_handler which pull
                function to call. Defaults to 'default'.

        Returns:
            not_pushed_ (dict): VQIP of water that cannot be pushed
        """
        if len(self.out_arcs) == 1:
            # If only one out_arc, just send the water down that
            if of_type is None:
                not_pushed_ = next(iter(self.out_arcs.values())).send_push_request(
                    vqip, tag=tag
                )
            elif any(
                [x in of_type for x, y in self.out_arcs_type.items() if len(y) > 0]
            ):
                not_pushed_ = next(iter(self.out_arcs.values())).send_push_request(
                    vqip, tag=tag
                )
            else:
                # No viable out arcs
                not_pushed_ = vqip
        else:
            # Push in proportion to connected by priority
            # Initialise pushed, deficit, connected, iter_
            not_pushed = vqip["volume"]
            not_pushed_ = self.copy_vqip(vqip)
            connected = self.get_connected(direction="push", of_type=of_type, tag=tag)
            iter_ = 0
            if not_pushed > connected["avail"]:
                # If more water than can be pushed, ignore preference and allocate all
                #   available based on capacity
                connected["priority"] = connected["avail"]
                connected["allocation"] = connected["capacity"]

            # Iterate over receiving nodes until sent
            while (
                (not_pushed > constants.FLOAT_ACCURACY)
                & (connected["avail"] > constants.FLOAT_ACCURACY)
                & (iter_ < constants.MAXITER)
            ):
                # Push to connected
                amount_to_push = min(connected["avail"], not_pushed)

                for key, allocation in connected["allocation"].items():
                    to_send = amount_to_push * allocation / connected["priority"]
                    to_send = self.v_change_vqip(not_pushed_, to_send)
                    reply = self.out_arcs[key].send_push_request(to_send, tag=tag)

                    sent = self.extract_vqip(to_send, reply)
                    not_pushed_ = self.extract_vqip(not_pushed_, sent)

                not_pushed = not_pushed_["volume"]
                connected = self.get_connected(
                    direction="push", of_type=of_type, tag=tag
                )
                iter_ += 1

            if iter_ == constants.MAXITER:
                print("Maxiter reached in {0} at {1}".format(self.name, self.t))

        return not_pushed_

    def check_basic(self, direction, vqip=None, of_type=None, tag="default"):
        """Generic function that conveys a pull or push check onwards to connected
        nodes. It is the default behaviour that treats a node like a junction.

        Args:
            direction (str): can be either 'pull' or 'push' to send checks to
                receiving or contributing nodes
            vqip (dict, optional): The VQIP to check. Defaults to None (if pulling
                this will return available water to pull, if pushing then available
                capacity to push).
            of_type (str or list) : optional, can be specified to send checks only
                to nodes of a given type (must be a subclass in nodes.py)
            tag (str, optional): optional message to direct query_handler which pull
                function to call. Defaults to 'default'.

        Returns:
            avail (dict): VQIP responses summed over all requests
        """
        f, arcs = self.get_direction_arcs(direction, of_type)

        # Iterate over arcs, updating total
        avail = self.empty_vqip()
        for arc in arcs:
            avail = self.sum_vqip(avail, getattr(arc, f)(tag=tag))

        if vqip is not None:
            avail = self.v_change_vqip(avail, min(avail["volume"], vqip["volume"]))

        return avail

    def pull_check_basic(self, vqip=None, of_type=None, tag="default"):
        """Default node check behaviour that treats a node like a junction. Water
        available to pull is just the water available to pull from upstream connected
        nodes.

        Args:
            vqip (dict, optional): VQIP from handler of amount to pull check
                (by default, only the 'volume' key is used). Defaults to None (which
                returns all availalbe water to pull).
            of_type (str or list) : optional, can be specified to send checks only
                to nodes of a given type (must be a subclass in nodes.py)
            tag (str, optional): optional message to direct query_handler which pull
                function to call. Defaults to 'default'.

        Returns:
            (dict): VQIP check response of upstream nodes
        """
        return self.check_basic("pull", vqip, of_type, tag)

    def push_check_basic(self, vqip=None, of_type=None, tag="default"):
        """Default node check behaviour that treats a node like a junction. Water
        available to push is just the water available to push to downstream connected
        nodes.

        Args:
            vqip (dict, optional): VQIP from handler of amount to push check.
                Defaults to None (which returns all available capacity to push).
            of_type (str or list) : optional, can be specified to send checks only
                to nodes of a given type (must be a subclass in nodes.py)
            tag (str, optional): optional message to direct query_handler which pull
                function to call. Defaults to 'default'.

        Returns:
            (dict): VQIP check response of downstream nodes
        """
        return self.check_basic("push", vqip, of_type, tag)

    def pull_set_deny(self, vqip):
        """Responds that no water is available to pull from a request.

        Args:
            vqip (dict): A VQIP amount of water requested (ignored)

        Returns:
            (dict): An empty VQIP indicated no water was pulled

        Raises:
            Message when called, since it would usually occur if a model is
            improperly connected
        """
        print("Attempted pull set from deny")
        return self.empty_vqip()

    def pull_check_deny(self, vqip=None):
        """Responds that no water is available to pull from a check.

        Args:
            vqip (dict): A VQIP amount of water requested (ignored)

        Returns:
            (dict): An empty VQIP indicated no water was pulled

        Raises:
            Message when called, since it would usually occur if a model is
            improperly connected
        """
        print("Attempted pull check from deny")
        return self.empty_vqip()

    def push_set_deny(self, vqip):
        """Responds that no water is available to push in a request.

        Args:
            vqip (dict): A VQIP amount of water to push

        Returns:
            vqip (dict): Returns the request indicating no water was pushed

        Raises:
            Message when called, since it would usually occur if a model is
            improperly connected
        """
        print("Attempted push set to deny")
        return vqip

    def push_check_deny(self, vqip=None):
        """Responds that no water is available to push in a check.

        Args:
            vqip (dict): A VQIP amount of water to push check (ignored)

        Returns:
            (dict): An empty VQIP indicated no capacity for pushes exists

        Raises:
            Message when called, since it would usually occur if a model is
            improperly connected
        """
        print("Attempted push check to deny")
        return self.empty_vqip()

    def push_check_accept(self, vqip=None):
        """Push check function that accepts all water.

        Args:
            vqip (dict, optional): A VQIP that has been pushed (ignored)

        Returns:
            (dict): VQIP or an unbounded capacity, indicating all water can be received
        """
        if not vqip:
            vqip = self.empty_vqip()
            vqip["volume"] = constants.UNBOUNDED_CAPACITY
        return vqip

    def get_data_input(self, var):
        """Read data from data_input_dict. Keys are tuples with the first entry as the
        variable to read and second entry the time.

        Args:
            var (str): Name of variable

        Returns:
            Data read
        """
        return self.data_input_dict[(var, self.t)]

    def end_timestep(self):
        """Empty function intended to be called at the end of every timestep.

        Subclasses will overwrite this functions.
        """
        pass

    def reinit(self):
        """Empty function to be written if reinitialisation capability is added."""
        pass


"""
    This is an attempt to generalise the behaviour of pull/push_distributed
    It doesn't yet work...

    def general_distribute(self, vqip, of_type = None, tag = 'default', direction =
        None):
        if direction == 'push':
            arcs = self.out_arcs
            arcs_type = self.out_arcs_type
            tracker = self.copy_vqip(vqip)
            requests = {x.name : lambda y : x.send_push_request(y, tag) for x in arcs.
                values()}
        elif direction == 'pull':
            arcs = self.in_arcs
            arcs_type = self.in_arcs_type
            tracker = self.empty_vqip()
            requests = {x.name : lambda y : x.send_pull_request(y, tag) for x in arcs.
                values()}
        else:
            print('No direction')

        if len(arcs) == 1:
            if (of_type == None) | any([x in of_type for x, y in arcs_type.items() if
                len(y) > 0]):
                arc = next(iter(arcs.keys()))
                return requests[arc](vqip)
            else:
                #No viable arcs
                return tracker

        connected = self.get_connected(direction = direction,
                                                                of_type = of_type,
                                                                tag = tag)

        iter_ = 0

        target = self.copy_vqip(vqip)
        #Iterate over sending nodes until deficit met
        while (((target['volume'] > constants.FLOAT_ACCURACY) &
                (connected['avail'] > constants.FLOAT_ACCURACY)) &
                (iter_ < constants.MAXITER)):

                amount = min(connected['avail'], target['volume']) #Deficit or amount
                    still to push
                replies = self.empty_vqip()

                for key, allocation in connected['allocation'].items():
                    to_request = amount * allocation / connected['priority']
                    to_request = self.v_change_vqip(target, to_request)
                    reply = requests[key](to_request)
                    replies = self.sum_vqip(replies, reply)

                if direction == 'pull':
                    target = self.extract_vqip(target, replies)
                elif direction == 'push':
                    target = replies

                connected = self.get_connected(direction = direction,
                                                                of_type = of_type,
                                                                tag = tag)
                iter_ += 1

                if iter_ == constants.MAXITER:
                    print('Maxiter reached')
        return target"""

# def replace(self, newnode):
#     #TODO - doesn't work because the bound methods (e.g., pull_set_handler) get
#     #associated with newnode and can't (as far as I can tell) be moved to self
#
#     """
#     Replace node with new node, maintaining any existing references to original node

#     Example
#     -------
#     print(my_node.name)
#     my_node.update(Node(name = 'new_node'))
#     print(my_node.name)
#     """


#     #You could use the below code to move the arcs_type references to arcs
#     #that are attached to this node. However, if you were creating a new
#     #subclass of Node then you would need to include this key in all
#     #arcs_type dictionaries in all nodes... which would get complicated
#     #probably safest to leave the arcs_type keys that are associated with
#     #this arc the same (for now...) -therefore - needs the same name

#     # for arc in self.in_arcs.values():
#     #     _ = arc.in_port.out_arcs_type[self.__class__.__name__].pop(arc.name)
#     #     arc.in_port.out_arcs_type[newnode.__class__.__name__][arc.name] = arc
#     # for arc in self.out_arcs.values():
#     #     _ = arc.out_port.in_arcs_type[self.__class__.__name__].pop(arc.name)
#     #     arc.out_port.in_arcs_type[newnode.__class__.__name__][arc.name] = arc


#     #Replace class
#     self.__class__ = newnode.__class__

#     #Replace object information (keeping some old information such as in_arcs)
#     for key in ['in_arcs',
#                 'out_arcs',
#                 'in_arcs_type',
#                 'out_arcs_type']:
#         newnode.__dict__[key] = self.__dict__[key]

#     self.__dict__.clear()
#     self.__dict__.update(newnode.__dict__)


NODES_REGISTRY: dict[str, type[Node]] = {Node.__name__: Node}



# -*- coding: utf-8 -*-
"""Created on Tue Oct 18 10:35:51 2022.

@author: Barney

"""
import os
import pytest
import unittest
import tempfile
import yaml
from unittest import TestCase, mock

from wsimod.arcs.arcs import Arc
from wsimod.nodes.land import Land
from wsimod.nodes.nodes import Node
from wsimod.nodes.sewer import Sewer
from wsimod.nodes.waste import Waste
from wsimod.orchestration.model import Model, to_datetime
import os


class MyTestClass(TestCase):
    def assertDictAlmostEqual(self, d1, d2, accuracy=19):
        """

        Args:
            d1:
            d2:
            accuracy:
        """
        for d in [d1, d2]:
            for key, item in d.items():
                d[key] = round(item, accuracy)
        self.assertDictEqual(d1, d2)

    def test_add_nodes(self):
        sewer = {"type_": "Sewer", "capacity": 0.04, "name": "my_sewer"}

        surface1 = {
            "type_": "ImperviousSurface",
            "surface": "urban",
            "area": 10,
            "pollutant_load": {"phosphate": 1e-7},
        }

        surface2 = {
            "type_": "PerviousSurface",
            "surface": "rural",
            "area": 100,
            "depth": 0.5,
            "pollutant_load": {"phosphate": 1e-7},
        }

        land = {"type_": "Land", "surfaces": [surface1, surface2], "name": "my_land"}
        my_model = Model()
        my_model.add_nodes([sewer, land])

        self.assertEqual(0.04, my_model.nodes["my_sewer"].sewer_tank.capacity)
        self.assertEqual(10, my_model.nodes["my_land"].get_surface("urban").area)

    def test_add_arcs(self):
        node = {"type_": "Node", "name": "my_node"}

        waste = {"type_": "Waste", "name": "my_waste"}

        surface1 = {
            "type_": "ImperviousSurface",
            "surface": "urban",
            "area": 10,
            "pollutant_load": {"phosphate": 1e-7},
        }

        surface2 = {
            "type_": "PerviousSurface",
            "surface": "rural",
            "area": 100,
            "depth": 0.5,
            "pollutant_load": {"phosphate": 1e-7},
        }

        land = {"type_": "Land", "surfaces": [surface1, surface2], "name": "my_land"}

        arc = {
            "in_port": "my_land",
            "out_port": "my_node",
            "name": "my_arc",
            "type_": "Arc",
            "capacity": 2,
            "preference": 2.5,
        }
        arc2 = {
            "in_port": "my_node",
            "out_port": "my_waste",
            "name": "my_arc2",
            "type_": "Arc",
        }
        my_model = Model()
        my_model.add_nodes([node, waste, land])
        my_model.add_arcs([arc, arc2])
        self.assertEqual(2, my_model.arcs["my_arc"].capacity)
        self.assertEqual(2.5, my_model.arcs["my_arc"].preference)

        self.assertEqual("my_land", my_model.arcs["my_arc"].in_port.name)
        self.assertEqual("my_node", my_model.arcs["my_arc"].out_port.name)

        self.assertEqual(
            0.4 * 0.5 * 100,
            my_model.arcs["my_arc"].in_port.get_surface("rural").capacity,
        )

    def test_add_ins_nodes(self):
        sewer = Sewer(**{"capacity": 0.04, "name": "my_sewer"})

        surface1 = {
            "type_": "ImperviousSurface",
            "surface": "urban",
            "area": 10,
            "pollutant_load": {"phosphate": 1e-7},
        }

        surface2 = {
            "type_": "PerviousSurface",
            "surface": "rural",
            "area": 100,
            "depth": 0.5,
            "pollutant_load": {"phosphate": 1e-7},
        }

        land = Land(**{"surfaces": [surface1, surface2], "name": "my_land"})
        my_model = Model()
        my_model.add_instantiated_nodes([sewer, land])

        self.assertEqual(0.04, my_model.nodes["my_sewer"].sewer_tank.capacity)
        self.assertEqual(10, my_model.nodes["my_land"].get_surface("urban").area)

    def test_add_ins_arcs(self):
        node = Node(**{"name": "my_node"})

        waste = Waste(**{"name": "my_waste"})

        surface1 = {
            "type_": "ImperviousSurface",
            "surface": "urban",
            "area": 10,
            "pollutant_load": {"phosphate": 1e-7},
        }

        surface2 = {
            "type_": "PerviousSurface",
            "surface": "rural",
            "area": 100,
            "depth": 0.5,
            "pollutant_load": {"phosphate": 1e-7},
        }

        land = Land(**{"surfaces": [surface1, surface2], "name": "my_land"})

        arc = Arc(
            **{
                "in_port": land,
                "out_port": node,
                "name": "my_arc",
                "capacity": 2,
                "preference": 2.5,
            }
        )
        arc2 = Arc(
            **{
                "in_port": node,
                "out_port": land,
                "name": "my_arc2",
            }
        )
        my_model = Model()
        my_model.add_instantiated_nodes([node, waste, land])
        my_model.add_instantiated_arcs([arc, arc2])
        self.assertEqual(2, my_model.arcs["my_arc"].capacity)
        self.assertEqual(2.5, my_model.arcs["my_arc"].preference)

        self.assertEqual("my_land", my_model.arcs["my_arc"].in_port.name)
        self.assertEqual("my_node", my_model.arcs["my_arc"].out_port.name)

        self.assertEqual(
            0.4 * 0.5 * 100,
            my_model.arcs["my_arc"].in_port.get_surface("rural").capacity,
        )

    def test_run(self):
        date = to_datetime("2000-01-01")
        dates = [date]
        land_inputs = {
            ("temperature", date): 10,
            ("precipitation", date): 0.01,
            ("et0", date): 0.002,
        }

        sewer = {"type_": "Sewer", "capacity": 0.05, "name": "my_sewer"}

        surface1 = {
            "type_": "ImperviousSurface",
            "surface": "urban",
            "area": 10,
            "pollutant_load": {"phosphate": 1e-7},
        }

        surface2 = {
            "type_": "PerviousSurface",
            "surface": "rural",
            "area": 100,
            "depth": 0.5,
            "pollutant_load": {"phosphate": 1e-7},
        }

        land = {
            "type_": "Land",
            "data_input_dict": land_inputs,
            "surfaces": [surface1, surface2],
            "name": "my_land",
        }

        gw = {
            "type_": "Groundwater",
            "area": 100,
            "capacity": 100,
            "name": "my_groundwater",
        }

        node = {"type_": "Node", "name": "my_river"}

        waste = {"type_": "Waste", "name": "my_outlet"}

        urban_drainage = {
            "type_": "Arc",
            "in_port": "my_land",
            "out_port": "my_sewer",
            "name": "urban_drainage",
        }

        percolation = {
            "type_": "Arc",
            "in_port": "my_land",
            "out_port": "my_groundwater",
            "name": "percolation",
        }

        runoff = {
            "type_": "Arc",
            "in_port": "my_land",
            "out_port": "my_river",
            "name": "runoff",
        }

        storm_outflow = {
            "type_": "Arc",
            "in_port": "my_sewer",
            "out_port": "my_river",
            "name": "storm_outflow",
        }

        baseflow = {
            "type_": "Arc",
            "in_port": "my_groundwater",
            "out_port": "my_river",
            "name": "baseflow",
        }

        catchment_outflow = {
            "type_": "Arc",
            "in_port": "my_river",
            "out_port": "my_outlet",
            "name": "catchment_outflow",
        }

        my_model = Model()
        my_model.dates = dates

        my_model.add_nodes([sewer, land, gw, node, waste])

        my_model.add_arcs(
            [
                urban_drainage,
                percolation,
                runoff,
                storm_outflow,
                baseflow,
                catchment_outflow,
            ]
        )

        flows, _, _, _ = my_model.run()
        self.assertEqual(0.05, flows[-1]["flow"])
        self.assertEqual(
            0.03, my_model.nodes["my_land"].get_surface("urban").storage["volume"]
        )

    def test_customise_orchestration(self):
        my_model = Model()
        my_model.load(
            os.path.join(os.getcwd(), "docs", "demo", "examples"),
            config_name="test_customise_orchestration_example.yaml",
        )
        revised_orchestration = [
            {"Groundwater": "infiltrate"},
            {"Sewer": "make_discharge"},
        ]
        self.assertListEqual(my_model.orchestration, revised_orchestration)

    def test_add_overrides_on_the_fly(self):
        overrides = {
            "nodes": {
                "my_groundwater": {
                    "name": "my_groundwater",
                    "type_": "Groundwater",
                    "infiltration_threshold": 200,
                    "infiltration_pct": 0.667,
                    "capacity": 1.43,
                    "area": 2.36,
                    "datum": 0.32,
                },
                "my_sewer": {
                    "name": "my_sewer",
                    "type_": "Sewer",
                    "pipe_timearea": {0: 0.5, 1: 0.5},
                },
            },
            "arcs": {
                "storm_outflow": {
                    "name": "storm_outflow",
                    "type_": "Arc",
                    "capacity": 0.5,
                }
            },
        }
        my_model = Model()
        my_model.load(
            os.path.join(os.getcwd(), "docs", "demo", "examples"),
            config_name="test_customise_orchestration_example.yaml",
        )
        my_model.add_overrides(overrides)

        self.assertEqual(200, my_model.nodes["my_groundwater"].infiltration_threshold)
        self.assertEqual(0.667, my_model.nodes["my_groundwater"].infiltration_pct)
        self.assertEqual(1.43, my_model.nodes["my_groundwater"].capacity)
        self.assertEqual(2.36, my_model.nodes["my_groundwater"].area)
        self.assertEqual(0.32, my_model.nodes["my_groundwater"].datum)
        self.assertEqual({0: 0.5, 1: 0.5}, my_model.nodes["my_sewer"].pipe_timearea)
        self.assertEqual(0.5, my_model.arcs["storm_outflow"].capacity)
        self.assertEqual(
            {
                "nodes": {"my_groundwater": {}, "my_sewer": {}},
                "arcs": {"storm_outflow": {}},
            },
            overrides,
        )

    def test_add_overrides_config(self):
        with open(
            os.path.join(
                os.getcwd(),
                "docs",
                "demo",
                "examples",
                "test_customise_orchestration_example.yaml",
            ),
            "r",
        ) as file:
            config = yaml.safe_load(file)
        config["overrides"] = {
            "nodes": {
                "my_groundwater": {
                    "name": "my_groundwater",
                    "type_": "Groundwater",
                    "infiltration_threshold": 200,
                    "infiltration_pct": 0.667,
                    "capacity": 1.43,
                    "area": 2.36,
                    "datum": 0.32,
                },
                "my_sewer": {
                    "name": "my_sewer",
                    "type_": "Sewer",
                    "pipe_timearea": {0: 0.5, 1: 0.5},
                },
            },
            "arcs": {
                "storm_outflow": {
                    "name": "storm_outflow",
                    "type_": "Arc",
                    "capacity": 0.5,
                }
            },
        }
        with tempfile.TemporaryDirectory() as tmp_path:
            with open(os.path.join(tmp_path, "temp.yaml"), "w") as file:
                yaml.dump(config, file)

            my_model = Model()
            my_model.load(tmp_path, config_name="temp.yaml")

        self.assertEqual(200, my_model.nodes["my_groundwater"].infiltration_threshold)
        self.assertEqual(0.667, my_model.nodes["my_groundwater"].infiltration_pct)
        self.assertEqual(1.43, my_model.nodes["my_groundwater"].capacity)
        self.assertEqual(2.36, my_model.nodes["my_groundwater"].area)
        self.assertEqual(0.32, my_model.nodes["my_groundwater"].datum)
        self.assertEqual({0: 0.5, 1: 0.5}, my_model.nodes["my_sewer"].pipe_timearea)
        self.assertEqual(0.5, my_model.arcs["storm_outflow"].capacity)


class TestLoadExtensionFiles:
    def test_load_extension_files_valid(self, tmp_path_factory):
        from wsimod.orchestration.model import load_extension_files

        with tmp_path_factory.mktemp("extensions") as tempdir:
            valid_file = os.path.join(tempdir, "valid_extension.py")
            with open(valid_file, "w") as f:
                f.write("def test_func(): pass")

            load_extension_files([valid_file])

    def test_load_extension_files_invalid_extension(self, tmp_path_factory):
        from wsimod.orchestration.model import load_extension_files

        with tmp_path_factory.mktemp("extensions") as tempdir:
            invalid_file = os.path.join(tempdir, "invalid_extension.txt")
            with open(invalid_file, "w") as f:
                f.write("This is a text file")

            with pytest.raises(ValueError, match="Only .py files are supported"):
                load_extension_files([invalid_file])

    def test_load_extension_files_nonexistent_file(self):
        from wsimod.orchestration.model import load_extension_files

        with pytest.raises(
            FileNotFoundError, match="File nonexistent_file.py does not exist"
        ):
            load_extension_files(["nonexistent_file.py"])

    def test_load_extension_files_import_error(self, tmp_path_factory):
        from wsimod.orchestration.model import load_extension_files

        with tmp_path_factory.mktemp("extensions") as tempdir:
            valid_file = os.path.join(tempdir, "valid_extension.py")
            with open(valid_file, "w") as f:
                f.write("raise ImportError")

            with pytest.raises(ImportError):
                load_extension_files([valid_file])


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""Created on Tue Oct 18 10:35:51 2022.

@author: Barney

"""

# import pytest
import unittest
from unittest import TestCase

from wsimod.core import constants
from wsimod.core.core import DecayObj, WSIObj


class MyTestClass(TestCase):
    def setUp(self):
        """"""
        constants.set_simple_pollutants()

    def test_empty(self):
        obj = WSIObj()
        self.assertDictEqual(
            {"volume": 0, "phosphate": 0, "temperature": 0}, obj.empty_vqip()
        )

    def test_copy(self):
        obj = WSIObj()
        d = {"volume": 10, "phosphate": 0.0001, "temperature": 15}
        self.assertDictEqual(d, obj.copy_vqip(d))

    def test_blend(self):
        obj = WSIObj()
        d1 = {"volume": 10, "phosphate": 0.0001, "temperature": 15}
        d2 = {"volume": 5, "phosphate": 0.00005, "temperature": 10}
        blend = {"volume": 15, "phosphate": 0.00125 / 15, "temperature": 40 / 3}
        self.assertDictEqual(blend, obj.blend_vqip(d1, d2))

    def test_sum(self):
        obj = WSIObj()
        d1 = {"volume": 10, "phosphate": 0.0001, "temperature": 15}
        d2 = {"volume": 5, "phosphate": 0.00005, "temperature": 10}
        blend = {"volume": 15, "phosphate": 0.0001 + 0.00005, "temperature": 40 / 3}
        self.assertDictEqual(blend, obj.sum_vqip(d1, d2))

    def test_to_total(self):
        obj = WSIObj()
        d = {"volume": 10, "phosphate": 0.0001, "temperature": 15}
        tot = {"volume": 10, "phosphate": 0.001, "temperature": 15}
        self.assertDictEqual(tot, obj.concentration_to_total(d))

    def test_to_concentration(self):
        obj = WSIObj()
        d = {"volume": 10, "phosphate": 0.0001, "temperature": 15}
        tot = {"volume": 10, "phosphate": 0.00001, "temperature": 15}
        self.assertDictEqual(tot, obj.total_to_concentration(d))

    def test_extract(self):
        obj = WSIObj()
        d1 = {"volume": 10, "phosphate": 0.0001, "temperature": 15}
        d2 = {"volume": 5, "phosphate": 0.00005, "temperature": 10}
        ext = {"volume": 5, "phosphate": 0.0001 - 0.00005, "temperature": 15}
        self.assertDictEqual(ext, obj.extract_vqip(d1, d2))

    def test_extract_c(self):
        obj = WSIObj()
        d1 = {"volume": 10, "phosphate": 0.0001, "temperature": 15}
        d2 = {"volume": 5, "phosphate": 0.00005, "temperature": 10}
        ext = {
            "volume": 5,
            "phosphate": (0.0001 * 10 - 0.00005 * 5) / 5,
            "temperature": 15,
        }
        self.assertDictEqual(ext, obj.extract_vqip_c(d1, d2))

    def test_distill(self):
        obj = WSIObj()
        d1 = {"volume": 10, "phosphate": 0.0001, "temperature": 15}
        v = 5
        ext = {"volume": 5, "phosphate": 0.0001, "temperature": 15}
        self.assertDictEqual(ext, obj.v_distill_vqip(d1, v))

    def test_distill_c(self):
        obj = WSIObj()
        d1 = {"volume": 10, "phosphate": 0.0001, "temperature": 15}
        v = 5
        ext = {"volume": 5, "phosphate": 0.0002, "temperature": 15}
        self.assertDictEqual(ext, obj.v_distill_vqip_c(d1, v))

    def test_vchange(self):
        obj = WSIObj()
        d1 = {"volume": 10, "phosphate": 0.0001, "temperature": 15}
        v = 5
        d = {"volume": 5, "phosphate": 0.00005, "temperature": 15}
        self.assertDictEqual(d, obj.v_change_vqip(d1, v))

    def test_vchange_c(self):
        obj = WSIObj()
        d1 = {"volume": 10, "phosphate": 0.0001, "temperature": 15}
        v = 5
        d = {"volume": 5, "phosphate": 0.0001, "temperature": 15}
        self.assertDictEqual(d, obj.v_change_vqip_c(d1, v))

    def test_ds(self):
        obj = WSIObj()
        d1 = {"volume": 10, "phosphate": 0.0001, "temperature": 15}
        d2 = {"volume": 5, "phosphate": 0.00005, "temperature": 10}
        ext = {"volume": 5, "phosphate": 0.00005, "temperature": 0}
        self.assertDictEqual(ext, obj.ds_vqip(d1, d2))

    def test_ds_c(self):
        obj = WSIObj()
        d1 = {"volume": 10, "phosphate": 0.0001, "temperature": 15}
        d2 = {"volume": 5, "phosphate": 0.00005, "temperature": 10}
        ext = {"volume": 5, "phosphate": 0.00075, "temperature": 0}
        self.assertDictEqual(ext, obj.ds_vqip_c(d1, d2))

    def test_compare(self):
        obj = WSIObj()
        d1 = {"volume": 10, "phosphate": 0.0001, "temperature": 15}
        d2 = {"volume": 5, "phosphate": 0.00005, "temperature": 10}
        self.assertFalse(obj.compare_vqip(d1, d2))
        self.assertTrue(obj.compare_vqip(d1, d1))

    def test_mass_balance(self):
        obj = WSIObj()
        d1 = {"volume": 100000, "phosphate": 0.001, "temperature": 15}

        d2 = {"volume": 99980.5, "phosphate": 0.00087, "temperature": 10}

        d3 = {"volume": 7.5, "phosphate": 0.00003, "temperature": 12}

        d4 = {"volume": 12, "phosphate": 0.0001, "temperature": 10}

        d5 = {"volume": 19.5, "phosphate": 0.0001 + 0.00003, "temperature": 0}
        obj.name = ""
        obj.mass_balance_in = [lambda: d1]
        obj.mass_balance_out = [lambda: d2]
        obj.mass_balance_ds = [lambda: d3, lambda: d4]
        in_, ds_, out_ = obj.mass_balance()

        self.assertDictEqual(in_, d1)
        self.assertDictEqual(out_, d2)
        self.assertDictEqual(ds_, d5)

    def test_generic_decay(self):
        class Do(DecayObj):
            """"""

            def __init__(self):
                self.parent = None
                DecayObj.__init__(self, decays=None)

        dobj = Do()
        temperature = 15
        decays = {"phosphate": {"constant": 0.001, "exponent": 1.005}}
        vq = {"volume": 10, "phosphate": 0.0001, "temperature": 15}
        diff = {
            "volume": 0,
            "phosphate": 0.0001 * 0.001 * 1.005 ** (15 - 20),
            "temperature": 0,
        }
        ext = {
            "volume": 10,
            "phosphate": 0.0001 - 0.0001 * 0.001 * 1.005 ** (15 - 20),
            "temperature": 15,
        }

        ext_, diff_ = dobj.generic_temperature_decay(vq, decays, temperature)

        ext_["phosphate"] = round(ext_["phosphate"], 15)
        ext["phosphate"] = round(ext["phosphate"], 15)
        diff["phosphate"] = round(diff["phosphate"], 15)
        diff_["phosphate"] = round(diff_["phosphate"], 15)

        self.assertDictEqual(ext, ext_)
        self.assertDictEqual(diff, diff_)

    def test_make_decay(self):
        class Do(DecayObj):
            """"""

            def __init__(self):
                self.parent = WSIObj()
                DecayObj.__init__(self, decays=None)

        dobj = Do()
        dobj.decays = {"phosphate": {"constant": 0.001, "exponent": 1.005}}
        dobj.data_input_object.data_input_dict = {("temperature", 1): 15}
        dobj.data_input_object.t = 1

        vq = {"volume": 10, "phosphate": 0.0001, "temperature": 15}
        diff = {
            "volume": 0,
            "phosphate": 0.0001 * 0.001 * 1.005 ** (15 - 20),
            "temperature": 0,
        }
        ext = {
            "volume": 10,
            "phosphate": 0.0001 - 0.0001 * 0.001 * 1.005 ** (15 - 20),
            "temperature": 15,
        }

        vq_ = dobj.make_decay(vq)
        td_ = dobj.total_decayed

        vq_["phosphate"] = round(vq_["phosphate"], 15)
        ext["phosphate"] = round(ext["phosphate"], 15)
        diff["phosphate"] = round(diff["phosphate"], 15)
        td_["phosphate"] = round(td_["phosphate"], 15)

        self.assertDictEqual(ext, vq_)
        self.assertDictEqual(diff, td_)

    def test_generic_decay_c(self):
        class Do(DecayObj):
            """"""

            def __init__(self):
                self.parent = None
                DecayObj.__init__(self, decays=None)

        dobj = Do()
        temperature = 15
        decays = {"phosphate": {"constant": 0.001, "exponent": 1.005}}
        vq = {"volume": 10, "phosphate": 0.0001, "temperature": 15}
        diff = {
            "volume": 0,
            "phosphate": 0.001 * 0.001 * 1.005 ** (15 - 20),
            "temperature": 0,
        }
        ext = {
            "volume": 10,
            "phosphate": 0.0001 - 0.0001 * 0.001 * 1.005 ** (15 - 20),
            "temperature": 15,
        }

        ext_, diff_ = dobj.generic_temperature_decay_c(vq, decays, temperature)

        ext_["phosphate"] = round(ext_["phosphate"], 15)
        ext["phosphate"] = round(ext["phosphate"], 15)
        diff["phosphate"] = round(diff["phosphate"], 15)
        diff_["phosphate"] = round(diff_["phosphate"], 15)

        self.assertDictEqual(ext, ext_)
        self.assertDictEqual(diff, diff_)


if __name__ == "__main__":
    unittest.main()



from typing import Optional

import pytest


@pytest.fixture
def temp_extension_registry():
    from wsimod.extensions import extensions_registry

    bkp = extensions_registry.copy()
    extensions_registry.clear()
    yield
    extensions_registry.clear()
    extensions_registry.update(bkp)


def test_register_node_patch(temp_extension_registry):
    from wsimod.extensions import extensions_registry, register_node_patch

    # Define a dummy function to patch a node method
    @register_node_patch("node_name", "method_name")
    def dummy_patch():
        print("Patched method")

    # Check if the patch is registered correctly
    assert extensions_registry[("node_name", "method_name", None, False)] == dummy_patch

    # Another function with other arguments
    @register_node_patch("node_name", "method_name", item="default", is_attr=True)
    def another_dummy_patch():
        print("Another patched method")

    # Check if this other patch is registered correctly
    assert (
        extensions_registry[("node_name", "method_name", "default", True)]
        == another_dummy_patch
    )


def test_apply_patches(temp_extension_registry):
    from wsimod.arcs.arcs import Arc
    from wsimod.extensions import (
        apply_patches,
        extensions_registry,
        register_node_patch,
    )
    from wsimod.nodes import Node
    from wsimod.orchestration.model import Model

    # Create a dummy model
    node = Node("dummy_node")
    node.dummy_arc = Arc("dummy_arc", in_port=node, out_port=node)
    model = Model()
    model.nodes[node.name] = node

    # 1. Patch a method
    @register_node_patch("dummy_node", "apply_overrides")
    def dummy_patch():
        pass

    # 2. Patch an attribute
    @register_node_patch("dummy_node", "t", is_attr=True)
    def another_dummy_patch(node):
        return f"A pathced attribute for {node.name}"

    # 3. Patch a method with an item
    @register_node_patch("dummy_node", "pull_set_handler", item="default")
    def yet_another_dummy_patch():
        pass

    # 4. Path a method of an attribute
    @register_node_patch("dummy_node", "dummy_arc.arc_mass_balance")
    def arc_dummy_patch():
        pass

    # Check if all patches are registered
    assert len(extensions_registry) == 4

    # Apply the patches
    apply_patches(model)

    # Verify that the patches are applied correctly
    assert (
        model.nodes[node.name].apply_overrides.__qualname__ == dummy_patch.__qualname__
    )
    assert (
        model.nodes[node.name]._patched_apply_overrides.__qualname__
        == "Node.apply_overrides"
    )
    assert model.nodes[node.name].t == another_dummy_patch(node)
    assert model.nodes[node.name]._patched_t == None
    assert (
        model.nodes[node.name].pull_set_handler["default"].__qualname__
        == yet_another_dummy_patch.__qualname__
    )
    assert (
        model.nodes[node.name].dummy_arc.arc_mass_balance.__qualname__
        == arc_dummy_patch.__qualname__
    )
    assert (
        model.nodes[node.name].dummy_arc._patched_arc_mass_balance.__qualname__
        == "Arc.arc_mass_balance"
    )


def assert_dict_almost_equal(d1: dict, d2: dict, tol: Optional[float] = None):
    """Check if two dictionaries are almost equal.

    Args:
        d1 (dict): The first dictionary.
        d2 (dict): The second dictionary.
        tol (float | None, optional): Relative tolerance. Defaults to 1e-6,
            `pytest.approx` default.
    """
    for key in d1.keys():
        assert d1[key] == pytest.approx(d2[key], rel=tol)


def test_path_method_with_reuse(temp_extension_registry):
    from wsimod.arcs.arcs import Arc
    from wsimod.extensions import apply_patches, register_node_patch
    from wsimod.nodes.storage import Reservoir
    from wsimod.orchestration.model import Model

    # Create a dummy model
    node = Reservoir(name="dummy_node", initial_storage=10, capacity=10)
    node.dummy_arc = Arc("dummy_arc", in_port=node, out_port=node)

    vq = node.pull_distributed({"volume": 5})
    assert_dict_almost_equal(vq, node.v_change_vqip(node.empty_vqip(), 5))

    model = Model()
    model.nodes[node.name] = node

    @register_node_patch("dummy_node", "pull_distributed")
    def new_pull_distributed(self, vqip, of_type=None, tag="default"):
        return self._patched_pull_distributed(vqip, of_type=["Node"], tag=tag)

    # Apply the patches
    apply_patches(model)

    # Check appropriate result
    assert node.tank.storage["volume"] == 5
    vq = model.nodes[node.name].pull_distributed({"volume": 5})
    assert_dict_almost_equal(vq, node.empty_vqip())
    assert node.tank.storage["volume"] == 5


def test_handler_extensions(temp_extension_registry):
    from wsimod.arcs.arcs import Arc
    from wsimod.extensions import apply_patches, register_node_patch
    from wsimod.nodes import Node
    from wsimod.orchestration.model import Model

    # Create a dummy model
    node = Node("dummy_node")
    node.dummy_arc = Arc("dummy_arc", in_port=node, out_port=node)
    model = Model()
    model.nodes[node.name] = node

    # 1. Patch a handler
    @register_node_patch("dummy_node", "pull_check_handler", item="default")
    def dummy_patch(self, *args, **kwargs):
        return "dummy_patch"

    # 2. Patch a handler with access to self
    @register_node_patch("dummy_node", "pull_set_handler", item="default")
    def dummy_patch(self, vqip, *args, **kwargs):
        return f"{self.name} - {vqip['volume']}"

    apply_patches(model)

    assert node.pull_check() == "dummy_patch"
    assert node.pull_set({"volume": 1}) == "dummy_node - 1"


def test_custom_class_from_file():
    """Test a custom class."""
    from pathlib import Path
    import yaml
    import tempfile

    from wsimod.nodes.nodes import NODES_REGISTRY
    from wsimod.orchestration.model import Model, to_datetime

    # Remove in case it was in there from previous test
    NODES_REGISTRY.pop("CustomNode", None)

    with tempfile.TemporaryDirectory() as temp_dir:
        config = {
            "nodes": {"node_name": {"type_": "CustomNode", "name": "node_name"}},
            "extensions": [str(Path(__file__).parent / "custom_class.py")],
        }

        with open(temp_dir + "/config.yml", "w") as f:
            yaml.dump(config, f)

        model = Model()
        model.load(temp_dir)
        assert model.nodes["node_name"].custom_attr == 1
        model.run(dates=[to_datetime("2000-01-01")])
        assert model.nodes["node_name"].custom_attr == 2

        model.save(temp_dir, "new_config.yml")

        # Remove the custom class from the registry to test loading it again
        del model
        NODES_REGISTRY.pop("CustomNode", None)

        model = Model()
        model.load(temp_dir, "new_config.yml")
        assert model.nodes["node_name"].custom_attr == 1
        model.run(dates=[to_datetime("2000-01-01")])
        assert model.nodes["node_name"].custom_attr == 2


def test_custom_class_on_the_fly():
    """Test a custom class."""

    import tempfile

    from wsimod.nodes.nodes import Node, NODES_REGISTRY
    from wsimod.orchestration.model import Model, to_datetime

    # Remove in case it was in there from previous test
    NODES_REGISTRY.pop("CustomNode", None)

    class CustomNode(Node):
        def __init__(self, name):
            super().__init__(name)
            self.custom_attr = 1

        def end_timestep(self):
            self.custom_attr += 1
            super().end_timestep()

    with tempfile.TemporaryDirectory() as temp_dir:
        model = Model()
        model.nodes["node_name"] = CustomNode("node_name")
        model.save(temp_dir)

        del model
        model = Model()
        model.load(temp_dir)
        assert model.nodes["node_name"].custom_attr == 1
        model.run(dates=[to_datetime("2000-01-01")])
        assert model.nodes["node_name"].custom_attr == 2



import ast
import os
from pathlib import Path
from typing import Any, Literal, Optional, Union

import pandas as pd
import yaml

from wsimod.core import constants


def evaluate_input_file(settings: Path) -> Literal["saved", "custom"]:
    """Decides what type of input file we are dealing with.

    "save" correspond to fully constructed models which have been saved, alongside
    any necessary data files. "custom" are input files constructed manually.

    Raises:
        ValueError: If the settings file do not exist.

    Return:
        If the input file is a saved model file or a custom input.
    """
    if settings.is_dir() or not settings.exists():
        raise ValueError(
            f"The settings file at {settings.absolute()} could not be found."
        )

    with settings.open("rb") as f:
        settings_ = yaml.safe_load(f)

    if set(["data", "inputs", "outputs"]).isdisjoint(settings_.keys()):
        return "saved"

    return "custom"


def validate_io_args(
    settings: Path, inputs: Optional[Path], outputs: Optional[Path]
) -> dict[str, Any]:
    """Validate the io arguments, including their definition in settings.

    This does not include validating the existance of data input files, which is done
    at a later stage.

    Args:
        settings (Path): The path to the file, in TOML format, containing all the
            configuration required for the simulation.
        inputs (Optional[Path]): Base directory for all input files. If present,
            overwrites value in the settings file.
        outputs (Optional[Path]): Base directory for all output files. If present,
            overwrites value in the settings file.

    Returns:
        dict[str, Any]: The loaded settings file with validated inputs and outputs.
    """
    if settings.is_dir() or not settings.exists():
        raise ValueError(
            f"The settings file at {settings.absolute()} could not be found."
        )

    with settings.open("rb") as f:
        settings_ = yaml.safe_load(f)

    # Valildate inputs folder
    settings_["inputs"] = _validate_input_dir(
        inputs if inputs else settings_.get("inputs", None), default=settings.parent
    )

    # Valildate outputs folder
    settings_["outputs"] = _validate_output_dir(
        outputs if outputs else settings_.get("outputs", None), default=settings.parent
    )

    return settings_


def _validate_input_dir(input_dir: Optional[Path], default: Path) -> Path:
    """Validates the directory of input files.

    If not provided, the default directory is used.

    Args:
        input_dir (Optional[Path]): The potential directory with the inputs.
        default (Path): Default input path if none provided.

    Raises:
        ValueError: If the inputs base directory is not actually a directory.

    Returns:
        Path: The validated path containing the inputs.
    """
    if not input_dir:
        return default.absolute()

    input_dir = Path(input_dir).absolute()
    if not input_dir.is_dir():
        raise ValueError(
            f"The inputs base directory at {input_dir} is not a directory."
        )
    return input_dir


def _validate_output_dir(output_dir: Optional[Path], default: Path) -> Path:
    """Validates the directory for output files.

    If not provided, the default path is used. If it does not exist, it is created.

    Args:
        output_dir (Optional[Path]): The potential directory for the outputs.
        default (Path): Defualt output path if none provided.

    Raises:
        ValueError: If a file with the same name already exist.

    Returns:
        Path: The validated path containing where outputs will be saved.
    """
    if not output_dir:
        return default.absolute()

    output_dir = Path(output_dir).absolute()
    if output_dir.exists() and not output_dir.is_dir():
        raise ValueError(f"A file at {output_dir} exists and is not a directory.")

    os.makedirs(output_dir, exist_ok=True)
    return output_dir


def load_data_files(
    data_settings: dict[str, Any], input_dir: Path
) -> dict[str, Union[pd.DataFrame, pd.Series, dict]]:
    """Reads the settings data section and reads the required data from files.

    Args:
        data_settings (dict[str, Any]): The data section of the settings file.
        input_dir (Path): The directory where input files are located.

    Returns:
        dict[str, Union[pd.DataFrame, pd.Series, dict]]: Loaded dataframe, series or
        dictionary following the instructions.
    """
    return {
        f"data:{key}": read_data(var, input_dir) for key, var in data_settings.items()
    }


def assign_data_to_settings(
    settings: dict[str, Any],
    data_settings: dict[str, Union[pd.DataFrame, pd.Series, dict]],
) -> dict[str, Any]:
    """Assigns the data files to the right variables in the settings dictionary.

    Search for data files to load is done recursively, walking through the whole
    settgins dictionary tree.

    Args:
        settings (dict[str, Any]): The settings dicitonary.
        input_dir (Path): The directory where input files are located.

    Returns:
        dict[str, Any]: A new settings dictionary where data files have been loaded.
    """
    loaded_settings: dict[str, Any] = {}

    for k, v in settings.items():
        if isinstance(v, dict):
            loaded_settings[k] = assign_data_to_settings(v, data_settings)
        elif isinstance(v, list):
            loaded_settings[k] = [
                assign_data_to_settings(item, data_settings) for item in v
            ]
        elif isinstance(v, str) and v.startswith("data:"):
            try:
                loaded_settings[k] = data_settings[v]
            except KeyError:
                raise ValueError(
                    f"{v} could not be found. Did you configure loading that data in"
                    " the data section of the settings file?"
                )
        else:
            loaded_settings[k] = v

    return loaded_settings


def read_data(
    instructions: dict[str, Any], inputs: Path
) -> Union[pd.DataFrame, pd.Series, dict]:
    """Uses the instructions to load tabular data.

    The instructions are a dictionary of options that define what file to load, how
    to load it and some simple manipulations to do to the loaded pandas Dataframe
    before returing it.

    The keys to control this proces are:

        filename: Filename of the data to load
        filter (optional): List of filters for the dataframe, each a dictionary in the
            form:
            where: column to filer
            is: value of that column
        scaling (optional): List of variable scaling, each a dictionary of the form:
            where: column to filer (optional)
            is: value of that column (optional)
            variable: name of the column to scale
            factor: unit conversion factor, as defined in `wsimod.core.constants`,
                eg. MM_TO_M
        format (optional): How the output should be provided. If format is `dict` then
            the output is provided as a dictonary, otherwise a Dataframe or a Series
            (if there is only 1 column) is output.
        index (optional): Column(s) to use as index.
        output (optional): Column to provide as output.
        options (optional): Options to pass to the `pandas.read_csv` function.

    The order in which operations are done is:

        read -> filter -> scale -> set_index -> select_output -> convert_format

    Only the `read` step will always happen. The others depend on the inputs.

    Args:
        instructions (str): A dictionary with instructions to load the data.
        inputs (Path): Base directory of inputs.

    Returns:
        Union[pd.DataFrame, pd.Series, dict]: Loaded dataframe, series or dictionary
        following the instructions.
    """
    filename = inputs / Path(instructions["filename"])
    options_: dict[str, Any] = process_options(instructions.get("options", ""))
    data = pd.read_csv(inputs / Path(filename), **options_)

    for filter in instructions.get("filters", []):
        data = data.loc[data[filter["where"]] == filter["is"]]

    for scaler in instructions.get("scaling", []):
        idx = data[scaler["where"]] == scaler["is"] if "is" in scaler else slice(None)
        factor = (
            getattr(constants, scaler["factor"])
            if isinstance(scaler["factor"], str)
            else scaler["factor"]
        )
        data.loc[idx, scaler["variable"]] *= factor

    if index := instructions.get("index", None):
        data = data.set_index(index)

    if output := instructions.get("output", None):
        data = data[output]

    if isinstance(data, pd.DataFrame) and len(data.columns) == 1:
        data = data.squeeze()

    if instructions.get("format", "") == "dict":
        return data.to_dict()

    return data


def process_options(options: str) -> dict[str, Any]:
    """Formats the options string as keyword arguments.

    >>> process_options("sep=' ',index_col='datetime'")
    {'sep': ' ', 'index_col': 'datetime'}

    Args:
        options (str): The strings with the arguments to process.

    Returns:
        dict[str, Any]: The dictionary with the processed keyword arguments.
    """
    if not options:
        return {}

    args = "f({})".format(options)
    tree = ast.parse(args)
    funccall = tree.body[0].value

    kwargs = {arg.arg: ast.literal_eval(arg.value) for arg in funccall.keywords}
    return kwargs



# -*- coding: utf-8 -*-
"""Created on Thu May 19 16:42:20 2022.

@author: barna
"""
from typing import Any, Dict

from wsimod.core import constants


class NutrientPool:
    """"""

    def __init__(
        self,
        fraction_dry_n_to_dissolved_inorganic=0.9,
        degrhpar={"N": 7 * 1e-5, "P": 7 * 1e-6},
        dishpar={"N": 7 * 1e-5, "P": 7 * 1e-6},
        minfpar={"N": 0.00013, "P": 0.000003},
        disfpar={"N": 0.000003, "P": 0.0000001},
        immobdpar={"N": 0.0056, "P": 0.2866},
        fraction_manure_to_dissolved_inorganic={"N": 0.5, "P": 0.1},
        fraction_residue_to_fast={"N": 0.1, "P": 0.1},
    ):
        """A class to track nutrient pools in a soil tank, intended to be initialised
        and called by GrowingSurfaces (see wsimod/nodes/land.py/GrowingSurface) and
        their subclasses. Contains five pools, which have a storage that tracks the mass
        of nutrients. Equations and parameters are based on HYPE.

        Args:
            fraction_dry_n_to_dissolved_inorganic (float, optional): fraction of dry
            nitrogen deposition going into the soil dissolved inorganic nitrogen pool,
            with the rest added to the fast pool. Defaults to 0.9. degrhpar (dict,
            optional): reference humus degradation rate (fraction of humus pool to fast
            pool). Defaults to {'N' : 7 * 1e-5, 'P' : 7 * 1e-6}. dishpar (dict,
            optional): reference humus dissolution rate (fraction of humus pool to
            dissolved organic pool). Defaults to {'N' : 7 * 1e-5, 'P' : 7 * 1e-6}.
            minfpar (dict, optional): reference fast pool mineralisation rate (fraction
            of fast pool to dissolved inorganic pool). Defaults to {'N' : 0.00013, 'P' :
            0.000003}. disfpar (dict, optional): reference fast pool dissolution rate
            (fraction of fast pool to dissolved organic pool). Defaults to {'N' :
            0.000003, 'P' : 0.0000001}. immobdpar (dict, optional): reference
            immobilisation rate (fraction of dissolved inorganic pool to fast pool).
            Defaults to {'N' : 0.0056, 'P' : 0.2866}.
            fraction_manure_to_dissolved_inorganic (dict, optional): fraction of
            nutrients from applied manure to dissolved inorganic pool, with the rest
            added to the fast pool. Defaults to {'N' : 0.5, 'P' : 0.1}.
            fraction_residue_to_fast (dict, optional): fraction of nutrients from
            residue to fast pool, with the rest added to the humus pool. Defaults to
            {'N' : 0.1, 'P' : 0.1}.

        Key assumptions:
             - Four nutrient pools are conceptualised for both nitrogen and phosphorus
                in soil, which includes humus pool, fast pool, dissolved inorganic pool,
                and dissolved organic pool. Humus and fast pool represent immobile pool
                of organic nutrients in the soil with slow and fast turnover,
                respectively. Dissolved inorganic and organic pool represent nutrients
                in dissolved phase in soil water (for phosphorus, dissolved organic pool
                might contain particulate phase). Given that phoshphorus can be adsorbed
                and attached to soil particles, an adsorbed inorganic pool is created
                specifically for phosphorus.
             - The major sources of nutrients to soil are conceptualised as
                - atmospheric deposition:
                    - dry deposition:
                        - for nitrogen, inorganic fraction of dry deposition is added to
                          the dissovled
                           inorganic pool, while the rest is added to the fast pool;
                        - for phosphorus, all is added to adsorbed inorganic pool.
                    - wet deposition: all is added to the dissolved inorganic pool.
                - fertilisers: all added to the dissolved inorganic pool.
                - manure: the inorganic fraction is added to the dissovled inorganic
                  pool, with
                    the rest added to the fast pool.
                - residue: the part with fast turnover is added to the fast pool, with
                  the rest
                    added to the humus pool.
             - Nutrient fluxes between these pools are simulated to represent the
               biochemical processes
                that can transform the nutrients between different forms. These
                processes include - degradation of humus pool to fast pool - dissolution
                of humus pool to dissovled organic pool - mineralisation of fast pool to
                dissolved inorganic pool - dissolution of fast pool to dissolved organic
                pool - immobilisation of dissolved inroganic pool to fast pool The rate
                of these processes are affected by the soil temperature and moisture
                conditions.
             - When soil erosion happens, a portion of both the adsorbed inorganic pool
               and humus pool
                for phosphorus will be eroded as well.

        Input data and parameter requirements:
             - fraction_dry_n_to_dissolved_inorganic,
               fraction_manure_to_dissolved_inorganic, fraction_residue_to_fast.
                _Units_: -, all should in [0-1]
             - degrhpar, dishpar, minfpar, disfpar, immobdpar.
                _Units_: -, all should in [0-1]
        """
        # TODO I don't think anyone will change most of these params... they could maybe
        # just be set here
        self.init_empty()

        # Assign parameters
        self.temperature_dependence_factor = 0
        self.soil_moisture_dependence_factor = 0

        self.fraction_manure_to_dissolved_inorganic = (
            fraction_manure_to_dissolved_inorganic
        )
        self.fraction_residue_to_fast = fraction_residue_to_fast
        self.fraction_dry_n_to_dissolved_inorganic = (
            fraction_dry_n_to_dissolved_inorganic
        )

        self.degrhpar = degrhpar
        self.dishpar = dishpar
        self.minfpar = minfpar
        self.disfpar = disfpar
        self.immobdpar = immobdpar

        self.fraction_manure_to_fast = None
        self.fraction_residue_to_humus = None
        self.fraction_dry_n_to_fast = None
        self.calculate_fraction_parameters()

        # Initialise different pools
        self.fast_pool = NutrientStore()
        self.humus_pool = NutrientStore()
        self.dissolved_inorganic_pool = NutrientStore()
        self.dissolved_organic_pool = NutrientStore()
        self.adsorbed_inorganic_pool = NutrientStore()
        self.pools = [
            self.fast_pool,
            self.humus_pool,
            self.dissolved_inorganic_pool,
            self.dissolved_organic_pool,
            self.adsorbed_inorganic_pool,
        ]

    def calculate_fraction_parameters(self):
        """Update fractions of nutrients input transformed into other forms in soil
        based on the input parameters
        Returns:
            (dict): fraction of manure to fast pool
            (dict): fraction of plant residue to humus pool
            (float): fraction of dry nitrogen deposition to fast pool
        """
        self.fraction_manure_to_fast = {
            x: 1 - self.fraction_manure_to_dissolved_inorganic[x]
            for x in constants.NUTRIENTS
        }
        self.fraction_residue_to_humus = {
            x: 1 - self.fraction_residue_to_fast[x] for x in constants.NUTRIENTS
        }
        self.fraction_dry_n_to_fast = 1 - self.fraction_dry_n_to_dissolved_inorganic

    def apply_overrides(self, overrides=Dict[str, Any]):
        """Override parameters.

        Enables a user to override any of the following parameters:
        eto_to_e, pore_depth.

        Args:
            overrides (Dict[str, Any]): Dict describing which parameters should
                be overridden (keys) and new values (values). Defaults to {}.
        """
        self.fraction_dry_n_to_dissolved_inorganic = overrides.pop(
            "fraction_dry_n_to_dissolved_inorganic",
            self.fraction_dry_n_to_dissolved_inorganic,
        )
        self.fraction_residue_to_fast.update(
            overrides.pop("fraction_residue_to_fast", {})
        )
        self.fraction_manure_to_dissolved_inorganic.update(
            overrides.pop("fraction_manure_to_dissolved_inorganic", {})
        )
        self.degrhpar.update(overrides.pop("degrhpar", {}))
        self.dishpar.update(overrides.pop("dishpar", {}))
        self.minfpar.update(overrides.pop("minfpar", {}))
        self.disfpar.update(overrides.pop("disfpar", {}))
        self.immobdpar.update(overrides.pop("immobdpar", {}))

        self.calculate_fraction_parameters()

    def init_empty(self):
        """Initialise an empty nutrient to be copied."""
        self.empty_nutrient = {x: 0 for x in constants.NUTRIENTS}

    def init_store(self):
        """Initialise an empty store to track nutrients."""
        self.init_empty()
        self.storage = self.get_empty_nutrient()

    def allocate_inorganic_irrigation(self, irrigation):
        """Assign inorganic irrigation, which is assumed to contain dissolved inorganic
        nutrients and thus updates that pool.

        Args:
            irrigation (dict): A dict that contains the amount of nutrients entering
                the nutrient pool via irrigation

        Returns:
            irrigation (dict): irrigation above, because no transformations take place
                (i.e., dissolved inorganic is what is received and goes straight into
                that pool)
        """
        # Update pool
        self.dissolved_inorganic_pool.receive(irrigation)
        return irrigation

    def allocate_organic_irrigation(self, irrigation):
        """Assign organic irrigation, which is assumed to contain dissolved organic
        nutrients and thus updates that pool.

        Args:
            irrigation (dict): A dict that contains the amount of nutrients entering
                the nutrient pool via irrigation

        Returns:
            irrigation (dict): irrigation above, because no transformations take place
                (i.e., dissolved organic is what is received and goes straight into that
                pool)
        """
        # Update pool
        self.dissolved_organic_pool.receive(irrigation)
        return irrigation

    def allocate_dry_deposition(self, deposition):
        """Assign dry deposition, which is assumed to go to both dissolved inorganic
        pool and fast pool (nitrogen) and the adsorbed pool (phosphorus).

        Args:
            deposition (dict): A dict that contains the amount of nutrients entering
                the nutrient pool via dry deposition

        Returns:
            (dict): A dict describing the amount of nutrients that enter the nutrient
                pool in a dissolved form (and thus need to be tracked by the soil water
                tank)
        """
        # Update pools
        self.fast_pool.storage["N"] += deposition["N"] * self.fraction_dry_n_to_fast
        self.dissolved_inorganic_pool.storage["N"] += (
            deposition["N"] * self.fraction_dry_n_to_dissolved_inorganic
        )
        self.adsorbed_inorganic_pool.storage["P"] += deposition["P"]
        return {
            "N": deposition["N"] * self.fraction_dry_n_to_dissolved_inorganic,
            "P": 0,
        }

    def allocate_wet_deposition(self, deposition):
        """Assign wet deposition, which is assumed to contain dissolved inorganic
        nutrients and thus updates that pool.

        Args:
            deposition (dict): A dict that contains the amount of nutrients entering
                the nutrient pool via wet deposition

        Returns:
            deposition (dict): deposition above, because no transformations take place
                (i.e., dissolved inorganic is what is received and goes straight into
                that pool)
        """
        # Update pool
        self.dissolved_inorganic_pool.receive(deposition)
        return deposition

    def allocate_manure(self, manure):
        """Assign manure, which is assumed to go to both dissolved inorganic pool and
        fast pool.

        Args:
            manure (dict): A dict that contains the amount of nutrients entering
                the nutrient pool via manure

        Returns:
            (dict): A dict describing the amount of nutrients that enter the nutrient
                pool in a dissolved form (and thus need to be tracked by the soil water
                tank)
        """
        # Assign a proportion of nutrients to the dissolved inorganic pool
        self.dissolved_inorganic_pool.receive(
            self.multiply_nutrients(manure, self.fraction_manure_to_dissolved_inorganic)
        )
        # Assign a proportion of nutrients to the fast pool
        self.fast_pool.receive(
            self.multiply_nutrients(manure, self.fraction_manure_to_fast)
        )
        return self.multiply_nutrients(
            manure, self.fraction_manure_to_dissolved_inorganic
        )

    def allocate_residue(self, residue):
        """Assign residue, which is assumed to go to both humus pool and fast pool.

        Args:
            residue (dict): A dict that contains the amount of nutrients entering
                the nutrient pool via residue

        Returns:
            (dict): A dict describing the amount of nutrients that enter the nutrient
                pool in a dissolved form (and thus need to be tracked by the soil water
                tank) - i.e., none because fast and humus pool are both solid
        """
        # Assign a proportion of nutrients to the humus pool
        self.humus_pool.receive(
            self.multiply_nutrients(residue, self.fraction_residue_to_humus)
        )
        # Assign a proportion of nutrients to the fast pool
        self.fast_pool.receive(
            self.multiply_nutrients(residue, self.fraction_residue_to_fast)
        )
        return self.empty_nutrient()

    def allocate_fertiliser(self, fertiliser):
        """Assign fertiliser, which is assumed to contain dissolved inorganic nutrients
        and thus updates that pool.

        Args:
            fertiliser (dict): A dict that contains the amount of nutrients entering
                the nutrient pool via fertiliser

        Returns:
            fertiliser (dict): fertiliser above, because no transformations take place
                (i.e., dissolved inorganic is what is received and goes straight into
                that pool)
        """
        self.dissolved_inorganic_pool.receive(fertiliser)
        return fertiliser

    def extract_dissolved(self, proportion):
        """Function to extract some amount of nutrients from all dissolved pools.

        Args:
            proportion (float): proportion of the dissolved nutrient pools to extract

        Returns:
            (dict): A dict of dicts, where the top level distinguishes between organic
                and inorganic nutrients, and the bottom level describes how much
                nutrients (i.e., N and P) have been extracted from those pools
        """
        # Extract from dissolved inorganic pool
        reply_di = self.dissolved_inorganic_pool.extract(
            {
                "N": self.dissolved_inorganic_pool.storage["N"] * proportion,
                "P": self.dissolved_inorganic_pool.storage["P"] * proportion,
            }
        )

        # Extract from dissolved organic pool
        reply_do = self.dissolved_organic_pool.extract(
            {
                "N": self.dissolved_organic_pool.storage["N"] * proportion,
                "P": self.dissolved_organic_pool.storage["P"] * proportion,
            }
        )
        return {"organic": reply_do, "inorganic": reply_di}

    def get_erodable_P(self):
        """Return total phosphorus that can be eroded (i.e., humus and adsorbed
        inorganic pools).

        Returns:
            (float): total phosphorus
        """
        return self.adsorbed_inorganic_pool.storage["P"] + self.humus_pool.storage["P"]

    def erode_P(self, amount_P):
        """Update humus and adsorbed inorganic pools to erode some amount. Removed in
        proportion to amount in both pools.

        Args:
            amount_P (float): Amount of phosphorus to be eroded

        Returns:
            (float): Amount of phosphorus eroded from the humus pool (float): Amount of
            phosphorus eroded from the adsorbed inorganic pool
        """
        # Calculate proportion of adsorbed to be eroded
        fraction_adsorbed = self.adsorbed_inorganic_pool.storage["P"] / (
            self.adsorbed_inorganic_pool.storage["P"] + self.humus_pool.storage["P"]
        )

        # Update nutrients in a dict holder
        request = self.get_empty_nutrient()

        # Update inorganic pool
        request["P"] = amount_P * fraction_adsorbed
        reply_adsorbed = self.adsorbed_inorganic_pool.extract(request)

        # Update humus pool
        request["P"] = amount_P * (1 - fraction_adsorbed)
        reply_humus = self.humus_pool.extract(request)

        return reply_humus["P"], reply_adsorbed["P"]

    def soil_pool_transformation(self):
        """Function to be called by a GrowingSurface that performs and tracks changes
        resulting from soil transformation processes.

        Returns:
            (float): increase in dissolved inorganic nutrients resulting from
                transformations (negative value indicates a decrease)
            (float): increase in dissolved organic nutrients resulting from
                transformations (negative value indicates a decrease)
        """
        # For mass balance purposes, assume fast is inorganic and humus is organic

        # Initialise tracking
        increase_in_dissolved_inorganic = self.get_empty_nutrient()
        increase_in_dissolved_organic = self.get_empty_nutrient()

        # Turnover of humus
        amount = self.temp_soil_process(self.degrhpar, self.humus_pool, self.fast_pool)
        # This is solid inorganic to solid organic... no tracking needed since solid
        # nutrients aren't tracked in mass balance of the surface soil water tank!

        # Dissolution of humus
        amount = self.temp_soil_process(
            self.dishpar, self.humus_pool, self.dissolved_organic_pool
        )
        increase_in_dissolved_organic = self.sum_nutrients(
            increase_in_dissolved_organic, amount
        )

        # Turnover of fast
        amount = self.temp_soil_process(
            self.minfpar, self.fast_pool, self.dissolved_inorganic_pool
        )
        increase_in_dissolved_inorganic = self.sum_nutrients(
            increase_in_dissolved_inorganic, amount
        )

        # Dissolution of fast
        amount = self.temp_soil_process(
            self.disfpar, self.fast_pool, self.dissolved_organic_pool
        )
        increase_in_dissolved_organic = self.sum_nutrients(
            increase_in_dissolved_organic, amount
        )

        # Immobilisation
        amount = self.temp_soil_process(
            self.immobdpar, self.dissolved_inorganic_pool, self.fast_pool
        )
        increase_in_dissolved_inorganic = self.subtract_nutrients(
            increase_in_dissolved_inorganic, amount
        )  # TODO will a negative value affect the consequent processes in growing
        # surface?

        return increase_in_dissolved_inorganic, increase_in_dissolved_organic

    def temp_soil_process(self, parameter, extract_pool, receive_pool):
        """Temperature function to take a parameter, calculate transformation, and
        remove nutrients from the extract pool and update the receive pool.

        Args:
            parameter (dict): A dict containing a parameter for each nutrient for the
            given process
                (units in per timestep)
            extract_pool (NutrientStore): The pool to extract from receive_pool
            (NutrientStore): The pool to receive extracted nutrients

        Returns:
            to_extract (dict): A dict containing the amount extracted of each nutrient
            (for mass
                balance)
        """
        # Initialise nutrients
        to_extract = self.get_empty_nutrient()
        for nutrient in constants.NUTRIENTS:
            # Calculate
            to_extract[nutrient] = (
                parameter[nutrient]
                * self.temperature_dependence_factor
                * self.soil_moisture_dependence_factor
                * extract_pool.storage[nutrient]
            )
        # Update pools
        to_extract = extract_pool.extract(to_extract)
        receive_pool.receive(to_extract)
        return to_extract

    def get_empty_nutrient(self):
        """An efficient way to get an empty nutrient.

        Returns:
            (dict): A dict containing 0 for each nutrient
        """
        return self.empty_nutrient.copy()

    def multiply_nutrients(self, nutrient, factor):
        """Multiply nutrients by factors.

        Args:
            nutrient (dict): Dict of nutrients to multiply factor (dict): Dict of
            factors to multiply for each nutrient

        Returns:
            (dict): Multiplied nutrients
        """
        return {x: nutrient[x] * factor[x] for x in constants.NUTRIENTS}

    def receive(self, nutrients):
        """Update nutrient store by amounts.

        Args:
            nutrients (dict): Amount of nutrients to update store by
        """
        # Increase storage
        for nutrient, amount in nutrients.items():
            self.storage[nutrient] += amount

    def sum_nutrients(self, n1, n2):
        """Sum two nutrients.

        Args:
            n1 (dict): Dict of nutrients n2 (dict): Dict of nutrients

        Returns:
            (dict): Summed nutrients
        """
        reply = self.get_empty_nutrient()
        for nutrient in constants.NUTRIENTS:
            reply[nutrient] = n1[nutrient] + n2[nutrient]
        return reply

    def subtract_nutrients(self, n1, n2):
        """Subtract two nutrients.

        Args:
            n1 (dict): Dict of nutrients to subtract from n2 (dict): Dict of nutrients
            to subtract

        Returns:
            (dict): subtracted nutrients
        """
        reply = self.get_empty_nutrient()
        for nutrient in constants.NUTRIENTS:
            reply[nutrient] = n1[nutrient] - n2[nutrient]
        return reply

    def extract(self, nutrients):
        """Remove nutrients from a store.

        Args:
            nutrients (dict): Dict of nutrients to remove from store

        Returns:
            (dict): amount of nutrients successfully removed
        """
        reply = self.get_empty_nutrient()
        for nutrient, amount in nutrients.items():
            reply[nutrient] = min(self.storage[nutrient], amount)
            self.storage[nutrient] -= reply[nutrient]

        return reply


class NutrientStore(NutrientPool):
    """"""

    def __init__(self):
        """Nutrient store, to be instantiated by a NutrientPool."""
        super().init_store()



# %% [markdown]
# # WSIMOD model demonstration - Oxford (.py)
#
# Note - this script can also be opened in interactive Python if you wanted to
# play around. On the GitHub it is in [docs/demo/scripts](https://github.com/barneydobson/wsi/blob/main/docs/demo/scripts/oxford_demo.py)
#
# 1. [Introduction](#we-will-cover-a-demo-wsimod-case-study)
#
# 2. [Data](#imports-and-forcing-data)
#
# 3. [Nodes](#create-nodes)
#
#     3.1 [Freshwater Treatment Works](#freshwater-treatment-works)
#
#     3.2 [Land](#land)
#
#     3.3 [Demand](#residential-demand)
#
#     3.4 [Reservoir](#reservoir)
#
#     3.5 [Distribution](#distribution)
#
#     3.6 [Wastewater Treatment Works](#wastewater-treatment-works)
#
#     3.7 [Sewers](#sewers)
#
#     3.8 [Groundwater](#groundwater)
#
#     3.9 [Node list](#create-a-nodelist)
#
# 4. [Arcs](#arcs)
#
#     4.1 [Arc parameters](#arc-parameters)
#
#     4.2 [Create the arcs](#create-arcs)
#
# 5. [Mapping](#mapping)
#
# 6. [Orchestration](#orchestration)
#
#     6.1 [Orchestrating an individual timestep](#orchestrating-an-individual-timestep)
#
#     6.2 [Ending a timestep](#ending-to-timestep)
#
# 7. [Model object](#model-object)
#
#     7.1 [Validation](#validation-plots)
# %% [markdown]
# ## We will cover a demo WSIMOD case study
#
# The glamorous town of Oxford will be our demo case study.
# Below, we will create these nodes and arcs, orchestrate them into a model, and run simulations.
#
# ![alt text](./../images/oxford.svg)
#
# Although GIS is pretty, the schematic below is a more accurate representation of what will be created.
# WSIMOD treats everything as a node or an arc.
#
# ![alt text](./../images/schematic.svg)
#

# %% [markdown]
# ## Imports and forcing data

# %% [markdown]
# Import packages
# %%

import os

import pandas as pd

from wsimod.arcs.arcs import Arc
from wsimod.core import constants
from wsimod.nodes.catchment import Catchment
from wsimod.nodes.demand import ResidentialDemand
from wsimod.nodes.land import Land
from wsimod.nodes.nodes import Node
from wsimod.nodes.sewer import Sewer
from wsimod.nodes.storage import Groundwater, Reservoir
from wsimod.nodes.waste import Waste
from wsimod.nodes.wtw import FWTW, WWTW
from wsimod.orchestration.model import Model

os.environ["USE_PYGEOS"] = "0"
import geopandas as gpd
from matplotlib import pyplot as plt
from shapely.geometry import LineString

# %% [markdown]
# Load input data
# %%

# Select the root path for the data folder. Use the appropriate value for your case.
data_folder = os.path.join(os.path.abspath(""), "docs", "demo", "data")

input_fid = os.path.join(data_folder, "processed", "timeseries_data.csv")
input_data = pd.read_csv(input_fid)
input_data.loc[input_data.variable == "flow", "value"] *= constants.M3_S_TO_M3_DT
input_data.loc[input_data.variable == "precipitation", "value"] *= constants.MM_TO_M
input_data.date = pd.to_datetime(input_data.date)
data_input_dict = input_data.set_index(["variable", "date"]).value.to_dict()
data_input_dict = (
    input_data.groupby("site")
    .apply(lambda x: x.set_index(["variable", "date"]).value.to_dict())
    .to_dict()
)
print(input_data.sample(10))

# %% [markdown]
# Input data is stored in dicts
# %%


print(data_input_dict["cherwell"][("boron", pd.to_datetime("2010-11-20"))])

# %% [markdown]
# We select dates that are available in the input data

# %%


dates = input_data.date.unique()
dates = dates[dates.argsort()]
dates = [pd.Timestamp(x) for x in dates]
print(dates[0:10])
# %% [markdown]
# We can specify the pollutants.
# In this example we choose based on what pollutants we have input data for.
# %%


constants.POLLUTANTS = input_data.variable.unique().tolist()
constants.POLLUTANTS.remove("flow")
constants.POLLUTANTS.remove("precipitation")
constants.POLLUTANTS.remove("et0")
constants.NON_ADDITIVE_POLLUTANTS = ["temperature"]
constants.ADDITIVE_POLLUTANTS = list(
    set(constants.POLLUTANTS).difference(constants.NON_ADDITIVE_POLLUTANTS)
)
constants.FLOAT_ACCURACY = 1e-11
print(constants.POLLUTANTS)


# %% [markdown]
# ## Create nodes

# %% [markdown]
# For [waste nodes](./../../../reference-other/#wsimod.nodes.waste.Waste),
# no parameters are needed, they are just the model outlet
# %%
thames_above_abingdon = Waste(name="thames_above_abingdon")

# %% [markdown]
# For junctions and abstraction locations, we can simply use the default
# [nodes](./../../../reference-nodes/#wsimod.nodes.nodes.Node)
# %%
farmoor_abstraction = Node(name="farmoor_abstraction")
evenlode_thames = Node(name="evenlode_thames")
cherwell_ray = Node(name="cherwell_ray")
cherwell_thames = Node(name="cherwell_thames")
thames_mixer = Node(name="thames_mixer")

# %% [markdown]
# For [catchment nodes](./../../../reference-other/#wsimod.nodes.catchment.Catchment),
# we only need to specify the input data (as a dictionary format).
# %%
evenlode = Catchment(name="evenlode", data_input_dict=data_input_dict["evenlode"])
thames = Catchment(name="thames", data_input_dict=data_input_dict["thames"])
ray = Catchment(name="ray", data_input_dict=data_input_dict["ray"])
cherwell = Catchment(name="cherwell", data_input_dict=data_input_dict["cherwell"])


# %% [markdown]
# We can see that, even though we provided mimimal information (name and input data) each node comes with many predefined functions.
# %%
print(dir(evenlode))

# %% [markdown]
# ### Freshwater treatment works
# Each type of node uses different parameters (see [API reference](./../../../../reference)).
# Below we create a [freshwater treatment works (FWTW)](./../../../reference-wtw/#wsimod.nodes.wtw.FWTW)

# %%

oxford_fwtw = FWTW(
    service_reservoir_storage_capacity=1e5,
    service_reservoir_storage_area=2e4,
    treatment_throughput_capacity=4.5e4,
    name="oxford_fwtw",
)

# %% [markdown]
# Each node type has different types of functionality available
# %%

print(dir(oxford_fwtw))

# %% [markdown]
# The FWTW node has a tank representing the service reservoirs, we can see that it has been initialised empty.

# %%

print(oxford_fwtw.service_reservoir_tank.storage)

# %% [markdown]
# If we try to pull water from the FWTW, it responds that there is no water to pull.

# %%

print(oxford_fwtw.pull_check({"volume": 10}))

# %% [markdown]
# If we add in some water, we see the pull check responds that water is available.

# %%

oxford_fwtw.service_reservoir_tank.storage["volume"] += 25
print(oxford_fwtw.pull_check({"volume": 10}))

# %% [markdown]
# When we set a pull request, we see that we successfully receive the water and the tank is updated.

# %%


reply = oxford_fwtw.pull_set({"volume": 10})
print(reply)
print(oxford_fwtw.service_reservoir_tank.storage)

# %% [markdown]
# ### Land
# We will now create a [land node](./../../../reference-land/#wsimod.nodes.land.Land),
# it is a bit involved so you might want to skip ahead to [demand](#Residential-demand),
# or check out the [land node tutorial](./../land_demo)
# %% [markdown]
# Data inputs are a single dictionary

# %%

land_inputs = data_input_dict["oxford_land"]

print(land_inputs[("precipitation", pd.to_datetime("2010-11-20"))])
print(land_inputs[("et0", pd.to_datetime("2010-11-20"))])
print(land_inputs[("temperature", pd.to_datetime("2009-10-15"))])

# %% [markdown]
# Assign some default pollutant deposition values (kg/m2/d)
# %%
pollutant_deposition = {
    "boron": 100e-10,
    "calcium": 70e-7,
    "chloride": 60e-10,
    "fluoride": 0.2e-7,
    "magnesium": 6e-7,
    "nitrate": 2e-9,
    "nitrogen": 4e-7,
    "potassium": 7e-7,
    "silicon": 7e-9,
    "sodium": 30e-9,
    "sulphate": 70e-7,
}
# %% [markdown]
# Create two surfaces as a list of dicts
# %%
surface = [
    {
        "type_": "PerviousSurface",
        "area": 2e7,
        "pollutant_load": pollutant_deposition,
        "surface": "rural",
        "field_capacity": 0.3,
        "depth": 0.5,
        "initial_storage": 2e7 * 0.4 * 0.5,
    },
    {
        "type_": "ImperviousSurface",
        "area": 1e7,
        "pollutant_load": pollutant_deposition,
        "surface": "urban",
        "initial_storage": 5e6,
    },
]

# %% [markdown]
# Create the land node from these surfaces and the input data
# %%
oxford_land = Land(surfaces=surface, name="oxford_land", data_input_dict=land_inputs)

# %% [markdown]
# We can see the land node has various tanks that have been initialised empty

# %%

print(oxford_land.surface_runoff.storage)
print(oxford_land.subsurface_runoff.storage)
print(oxford_land.percolation.storage)
# %% [markdown]
# We can see the surfaces have also been initialised, although they are not empty because we provided 'initial_storage' parameters.

# %%
rural_surface = oxford_land.get_surface("rural")
urban_surface = oxford_land.get_surface("urban")
print("{0}-{1}".format("rural", rural_surface.storage))
print("{0}-{1}".format("urban", urban_surface.storage))

# %% [markdown]
# We can run a timestep of the land node with the 'run' command

# %%
oxford_land.t = pd.to_datetime("2012-12-22")

oxford_land.run()

# %% [markdown]
# We can see that the land and surface tanks have been updated

# %%
print(oxford_land.surface_runoff.storage)
print(oxford_land.subsurface_runoff.storage)
print(oxford_land.percolation.storage)

print("{0}-{1}".format("rural", rural_surface.storage))
print("{0}-{1}".format("urban", urban_surface.storage))

# %% [markdown]
# ### Residential demand
# The [residential demand](./../../../reference-other/#wsimod.nodes.demand.ResidentialDemand)
# node requires population, per capita demand and a pollutant_load dictionary that defines
# how much (weight in kg) pollution is generated per person per day.

# %%
oxford = ResidentialDemand(
    name="oxford",
    population=2e5,
    per_capita=0.15,
    pollutant_load={
        "boron": 500 * constants.UG_L_TO_KG_M3 * 0.15,
        "calcium": 150 * constants.MG_L_TO_KG_M3 * 0.15,
        "chloride": 180 * constants.MG_L_TO_KG_M3 * 0.15,
        "fluoride": 0.4 * constants.MG_L_TO_KG_M3 * 0.15,
        "magnesium": 30 * constants.MG_L_TO_KG_M3 * 0.15,
        "nitrate": 60 * constants.MG_L_TO_KG_M3 * 0.15,
        "nitrogen": 50 * constants.MG_L_TO_KG_M3 * 0.15,
        "potassium": 30 * constants.MG_L_TO_KG_M3 * 0.15,
        "silicon": 20 * constants.MG_L_TO_KG_M3 * 0.15,
        "sodium": 200 * constants.MG_L_TO_KG_M3 * 0.15,
        "sulphate": 250 * constants.MG_L_TO_KG_M3 * 0.15,
        "temperature": 14,
    },
    data_input_dict=land_inputs,
)
# pollutant_load calculated based on expected effluent at WWTW


# %% [markdown]
# ### Reservoir
# A [reservoir node](./../../../reference-storage/#wsimod.nodes.storage.Reservoir)
# is used to make abstractions from rivers and supply FWTWs

# %%
farmoor = Reservoir(
    name="farmoor", capacity=1e7, initial_storage=1e7, area=1.5e6, datum=62
)

# %% [markdown]
# ### Distribution
# We use a generic [Node](./../../../reference-nodes/#wsimod.nodes.nodes.Node) as a junction to represent the distribution network between the FWTW and households

# %%

distribution = Node(name="oxford_distribution")

# %% [markdown]
# ### Wastewater treatment works
# [Wastewater treatment works (WWTW)](./../../../reference-wtw/#wsimod.nodes.wtw.WWTW)
# are nodes that can store sewage water temporarily in storm tanks, and reduce the pollution
# amounts in water before releasing them onwards to rivers.

# %%

oxford_wwtw = WWTW(
    stormwater_storage_capacity=2e4,
    stormwater_storage_area=2e4,
    treatment_throughput_capacity=5e4,
    name="oxford_wwtw",
)

# %% [markdown]
# ### Sewers
# [Sewer nodes](./../../../reference-sewer/#wsimod.nodes.sewer.Sewer) enable water to transition
# between households and WWTWs, and between impervious surfaces and rivers or WWTWs.
# They use a timearea diagram to represent travel time, which assigns a specified percentage
# of water to take a specified duration to pass through the sewer node.
# %%
combined_sewer = Sewer(
    capacity=4e6, pipe_timearea={0: 0.8, 1: 0.15, 2: 0.05}, name="combined_sewer"
)

# %% [markdown]
# ### Groundwater
# [Groundwater nodes](./../../../reference-storage/#wsimod.nodes.storage.Groundwater)
# implement a simple residence time to determine baseflow

# %%

gw = Groundwater(capacity=3.2e9, area=3.2e8, name="gw", residence_time=20)


# %% [markdown]
# ### Create a nodelist
# To keep all the nodes in one place, we put them into a list
# %%
nodelist = [
    thames_above_abingdon,
    evenlode,
    thames,
    ray,
    cherwell,
    oxford,
    distribution,
    farmoor,
    oxford_fwtw,
    oxford_wwtw,
    combined_sewer,
    oxford_land,
    gw,
    farmoor_abstraction,
    evenlode_thames,
    cherwell_ray,
    cherwell_thames,
    thames_mixer,
]

print(nodelist)

# %% [markdown]
# ## Arcs
# [Arcs](./../../../reference-arc/#wsimod.arcs.arcs.Arc) link nodes.
# An example arc is the link between a FWTW and the distribution node
# %%

# Standard simple arcs
fwtw_to_distribution = Arc(
    in_port=oxford_fwtw, out_port=distribution, name="fwtw_to_distribution"
)
print(fwtw_to_distribution)

# %% [markdown]
# As with nodes, even though we only gave it a few parameters, the arc comes with a lot built in
# %%
print(dir(fwtw_to_distribution))

# %% [markdown]
# We can see that the arc links the two nodes
# %%


print(fwtw_to_distribution.in_port)


# %%


print(fwtw_to_distribution.out_port)

# %% [markdown]
# And that it has updated the nodes that it is connecting.
# %%


print(oxford_fwtw.out_arcs)


# %%


print(distribution.in_arcs)

# %% [markdown]
# We use arcs to send checks and requests..
# %%


print(fwtw_to_distribution.send_pull_check({"volume": 20}))


# %%


reply = fwtw_to_distribution.send_pull_request({"volume": 20})
print(reply)

# %% [markdown]
# They convey this information to the nodes that they connect to, which update their state variables
# %%


print(oxford_fwtw.service_reservoir_tank.storage)

# %% [markdown]
# In turn, the arcs update their own state variables.
# %%


print(fwtw_to_distribution.flow_in)
print(fwtw_to_distribution.flow_out)

# %% [markdown]
# ## Arc parameters
# Besides the in/out ports and names, arcs can have a parameter for their capacity, to limit the flow that may pass through it each timestep.
# A typical example would be on river abstractions to a reservoir
# %%
abstraction_to_farmoor = Arc(
    in_port=farmoor_abstraction,
    out_port=farmoor,
    name="abstraction_to_farmoor",
    capacity=5e4,
)

# %% [markdown]
# A bit more sophisticated is the 'preference' parameter.
# We use preference to express where we would prefer the model to send water.
# In this example, the sewer can send water to both the treatment plant and directly into the river.
# Of course we would always to prefer to send water to the plant, so we give it a very high preference.
# Discharging into the river should only be done if there is no capacity left at the WWTW, so we give the arc a very low preference.
# %%
sewer_to_wwtw = Arc(
    in_port=combined_sewer, out_port=oxford_wwtw, preference=1e10, name="sewer_to_wwtw"
)
sewer_overflow = Arc(
    in_port=combined_sewer,
    out_port=thames_mixer,
    preference=1e-10,
    name="sewer_overflow",
)


# %% [markdown]
# ## Create arcs
# Arcs are a bit less interesting than nodes because they generally don't capture complicated physical behaviours.
# So we just create all of them below.
# %%


evenlode_to_thames = Arc(
    in_port=evenlode, out_port=evenlode_thames, name="evenlode_to_thames"
)

thames_to_thames = Arc(
    in_port=thames, out_port=evenlode_thames, name="thames_to_thames"
)

ray_to_cherwell = Arc(in_port=ray, out_port=cherwell_ray, name="ray_to_cherwell")

cherwell_to_cherwell = Arc(
    in_port=cherwell, out_port=cherwell_ray, name="cherwell_to_cherwell"
)

thames_to_farmoor = Arc(
    in_port=evenlode_thames, out_port=farmoor_abstraction, name="thames_to_farmoor"
)

farmoor_to_mixer = Arc(
    in_port=farmoor_abstraction, out_port=thames_mixer, name="farmoor_to_mixer"
)

cherwell_to_mixer = Arc(
    in_port=cherwell_ray, out_port=thames_mixer, name="cherwell_to_mixer"
)

wwtw_to_mixer = Arc(in_port=oxford_wwtw, out_port=thames_mixer, name="wwtw_to_mixer")

mixer_to_waste = Arc(
    in_port=thames_mixer, out_port=thames_above_abingdon, name="mixer_to_waste"
)

distribution_to_demand = Arc(
    in_port=distribution, out_port=oxford, name="distribution_to_demand"
)

reservoir_to_fwtw = Arc(in_port=farmoor, out_port=oxford_fwtw, name="reservoir_to_fwtw")

fwtw_to_sewer = Arc(in_port=oxford_fwtw, out_port=combined_sewer, name="fwtw_to_sewer")

demand_to_sewer = Arc(in_port=oxford, out_port=combined_sewer, name="demand_to_sewer")

land_to_sewer = Arc(in_port=oxford_land, out_port=combined_sewer, name="land_to_sewer")

land_to_gw = Arc(in_port=oxford_land, out_port=gw, name="land_to_gw")

garden_to_gw = Arc(in_port=oxford, out_port=gw, name="garden_to_gw")

gw_to_mixer = Arc(in_port=gw, out_port=thames_mixer, name="gw_to_mixer")

# %% [markdown]
# Again, we keep all the arcs in a tidy list together.

# %%
arclist = [
    evenlode_to_thames,
    thames_to_thames,
    ray_to_cherwell,
    cherwell_to_cherwell,
    thames_to_farmoor,
    farmoor_to_mixer,
    cherwell_to_mixer,
    wwtw_to_mixer,
    sewer_overflow,
    mixer_to_waste,
    abstraction_to_farmoor,
    distribution_to_demand,
    demand_to_sewer,
    land_to_sewer,
    sewer_to_wwtw,
    fwtw_to_sewer,
    fwtw_to_distribution,
    reservoir_to_fwtw,
    land_to_gw,
    garden_to_gw,
    gw_to_mixer,
]


# %% [markdown]
# ## Mapping
# Remember, WSIMOD is an integrated model.
# Because it covers so many different things, it is very easy to make mistakes.
# Thus it is always good practice to plot your data!
#
# Below we load the node location data and create arcs from the information in the arclist.
# %%

location_fn = os.path.join(data_folder, "raw", "points_locations.geojson")
nodes_gdf = gpd.read_file(location_fn).set_index("name")
arcs_gdf = []

for arc in arclist:
    arcs_gdf.append(
        {
            "name": arc.name,
            "geometry": LineString(
                [
                    nodes_gdf.loc[arc.in_port.name, "geometry"],
                    nodes_gdf.loc[arc.out_port.name, "geometry"],
                ]
            ),
        }
    )

arcs_gdf = gpd.GeoDataFrame(arcs_gdf, crs=nodes_gdf.crs)
# %% [markdown]
# Because we converted the information as GeoDataFrames, we can simply plot them below
# %%

f, ax = plt.subplots()
arcs_gdf.plot(ax=ax)
nodes_gdf.plot(color="r", ax=ax, zorder=10)

# %% [markdown]
# ## Orchestration
# Orchestration is making the simulation happen by calling functions in the nodes.
# These functions simulate physical behaviour within the node, and cause pulls/pushes to happen which in turn triggers physical behaviour in other nodes.
#
# ### Orchestrating an individual timestep
#
# We will start below by manually orchestrating a single timestep.
#
# We start by setting the date, so that every node knows what forcing data to read for this timestep.

# %%
date = dates[0]

for node in nodelist:
    node.t = date

print(date)
print(oxford_fwtw.t)

# %% [markdown]
# We can see the service reservoirs are empty but the supply reservoir is not!
# %%

print(oxford_fwtw.service_reservoir_tank.storage)
print(farmoor.tank.storage)
# %% [markdown]
# If we call the FWTW's treat_water function it will pull water from the supply reservoir and update its service reservoirs
# %%

oxford_fwtw.treat_water()

print(oxford_fwtw.service_reservoir_tank.storage)
print(farmoor.tank.storage)

# %% [markdown]
# This information is tracked in the arcs that enter the FWTW
# %%


print(oxford_fwtw.in_arcs)


# %%


print(reservoir_to_fwtw.flow_in)
print(reservoir_to_fwtw.flow_out)

# %% [markdown]
# Although none of that water has yet entered the distribution network (only some small flow from the earlier demonstration)
# %%


print(fwtw_to_distribution.flow_in)

# %% [markdown]
# That is because no water consumption demand had yet been generated.
#
# If we call the demand node's create_demand function we see that the distribution arc becomes utilised.
# %%

oxford.create_demand()
print(fwtw_to_distribution.flow_in)

# %% [markdown]
# We also see that this gets pushed onwards into the sewer system

# %%
print(demand_to_sewer.flow_in)


# %% [markdown]
# Many nodes have functions intended to be called during orchestration.
# These functions are described in the documentation.
# For example, we see in the [Land node](./../../../reference-land/#wsimod.nodes.land.Land) API reference that the 'run' function is intended to be called from orchestration.
# %%
oxford_land.run()

# %% [markdown]
# Below we call the functions for other nodes
# %%

# Discharge GW
gw.distribute()

# Discharge sewers (pushed to other sewers or WWTW)
combined_sewer.make_discharge()

# Run WWTW model
oxford_wwtw.calculate_discharge()

# Make abstractions
farmoor.make_abstractions()

# Discharge WW
oxford_wwtw.make_discharge()

# Route catchments
evenlode.route()
thames.route()
ray.route()
cherwell.route()

# %% [markdown]
# ## Ending a timestep
# Because mistakes happen, it is essential to carry out mass balance testing.
# Each node has a mass balance function that can be called.
# We see a mass balance violation resulting from the demonstration with the FWTW earlier.
# %%

for node in nodelist:
    in_, ds_, out_ = node.node_mass_balance()


# %% [markdown]
# We should also call the end_timestep function in nodes and arcs.
# This is important for mass balance testing and capturing the behaviour of some dynamic processes in nodes.

# %%

for node in nodelist:
    node.end_timestep()

for arc in arclist:
    arc.end_timestep()

# %% [markdown]
# ## Model object
# Of course it would be a massive pain to manually orchestrate every timestep.
# So instead we store node and arc information in a [model object](./../../../reference-model/#wsimod.orchestration.model.Model)
# that will do the orchestration for us.
#
# Because we have already created the nodes/arcs above, we simply need to add the instantiated lists above.

# %%

my_model = Model()
my_model.add_instantiated_nodes(nodelist)
my_model.add_instantiated_arcs(arclist)
my_model.dates = dates

# %% [markdown]
# The model object lets us reinitialise the nodes/arcs, and run all of the orchestration with a 'run' function.
# %%

my_model.reinit()

flows, _, _, _ = my_model.run()

# %% [markdown]
# The model outputs flows as a dictionary which can be converted to a dataframe
# %%

flows = pd.DataFrame(flows)

print(flows.sample(10))

# %% [markdown]
# ## Validation plots
# Of course we shouldn't trust a model without proof, and this is doubly true for an integrated model whose errors may easily propagate.
#
# Thankfully we have rich in-river water quality sampling in Oxford that we can use for validation
#
# We load and format this data for dates and pollutants that overlap with what we have simulated
# %%

mixer_val_df = pd.read_csv(os.path.join(data_folder, "raw", "mixer_wims_val.csv"))
mixer_val_df.date = pd.to_datetime(mixer_val_df.date)
mixer_val_df = mixer_val_df.loc[mixer_val_df.date.isin(dates)]
val_pollutants = mixer_val_df.variable.unique()
mixer_val_df = mixer_val_df.pivot(index="date", columns="variable", values="value")

# %% [markdown]
# We get rid of the first day of flows because our tanks were initialised empty and will not be informative

# %%
flows = flows.loc[flows.time != dates[0]]

# %% [markdown]
# We convert our flows, which are simulated in kg/d into a concentration value (kg/m3/d).
# %%
flows_plot = flows.copy()

# Convert to KG/M3
for pol in set(val_pollutants):
    if pol != "temperature":
        flows_plot[pol] /= flows_plot.flow
# %% [markdown]
# We pick the model outlet as a validation location and make a pretty plot - fantastic!
# %%

plot_arc = "mixer_to_waste"
f, axs = plt.subplots(val_pollutants.size, 1)
for pol, ax in zip(val_pollutants, axs):
    ax.plot(
        flows_plot.loc[flows_plot.arc == plot_arc, [pol, "time"]].set_index("time"),
        color="b",
        label="simulation",
    )
    ax.plot(mixer_val_df[pol], ls="", marker="o", color="r", label="spot sample")
    ax.set_ylabel(pol)
plt.legend()



# -*- coding: utf-8 -*-
"""Created on Mon Nov 15 14:20:36 2021.

@author: bdobson

Converted to totals on 2022-05-03
"""
from wsimod.core import constants
from wsimod.nodes.nodes import Node


class Catchment(Node):
    """"""

    def __init__(
        self,
        name,
        data_input_dict={},
    ):
        """Node that reads input data to create VQIPs that are pushed downstream and
        tracks abstractions made from the node, adjusting pushes accordingly.

        Args:
            name (str): Node name
            data_input_dict (dict, optional): Dictionary of data inputs relevant for
                the node. Keys are tuples where first value is the name of the
                variable to read from the dict and the second value is the time.
                Defaults to {}.

        Functions intended to call in orchestration:
            route

        Key assumptions:
            - Flows from `Catchment` nodes are simply read from data, thus
                assumed to be unresponsive to wider water cycle changes.

        Input data and parameter requirements:
            - Flow data in the `data_input_dict` at the model timestep.
                _Units_: cubic metres/timestep
            - Values for each variable defined in `constants.POLLUTANTS` also
                stored in `data_input_dict` at the model timestep.
                _Units_: kg/m3/timestep (additive pollutants)
        """
        # Update args
        super().__init__(name)
        self.data_input_dict = data_input_dict

        # Update handlers
        self.pull_set_handler["default"] = self.pull_set_abstraction
        self.pull_check_handler["default"] = self.pull_check_abstraction
        self.push_set_handler["default"] = self.push_set_deny
        self.push_check_handler["default"] = self.push_set_deny
        self.unrouted_water = self.empty_vqip()
        # Mass balance
        self.mass_balance_in.append(lambda: self.get_flow())
        self.mass_balance_out.append(lambda: self.unrouted_water)
        self.end_timestep = self.end_timestep_

    def get_flow(self):
        """Read volume data, read pollutant data, convert additibve pollutants from
        kg/m3 to kg.

        Returns:
            vqip (dict): Return read data as a VQIP
        """
        # TODO (if used) - note that if flow is < float accuracy then it won't
        # get pushed, and the pollutants will 'disappear', causing a mass balance error
        vqip = {"volume": self.data_input_dict[("flow", self.t)]}
        for pollutant in constants.POLLUTANTS:
            vqip[pollutant] = self.data_input_dict[(pollutant, self.t)]
        for pollutant in constants.ADDITIVE_POLLUTANTS:
            vqip[pollutant] *= vqip["volume"]

        return vqip

    def route(self):
        """Send any water that has not already been abstracted downstream."""
        # Get amount of water
        avail = self.get_avail()
        # Route excess flow onwards
        reply = self.push_distributed(avail, of_type=["Node", "River", "Waste"])
        self.unrouted_water = self.sum_vqip(self.unrouted_water, reply)
        if reply["volume"] > constants.FLOAT_ACCURACY:
            pass
            # print('Catchment unable to route')

    def get_avail(self):
        """Water available for abstraction (Read data and subtract pre-existing
        abstractions).

        Returns:
            avail (dict): A VQIP of water available for abstraction
        """
        # Get available vqip
        avail = self.get_flow()

        # Remove abstractions already made
        for name, arc in self.out_arcs.items():
            avail = self.v_change_vqip(avail, avail["volume"] - arc.vqip_in["volume"])

        return avail

    def pull_check_abstraction(self, vqip=None):
        """Check wrapper for get_avail that updates response if VQIP is given.

        Args:
            vqip (dict, optional): A VQIP that is compared with get_avail and the
                minimum is returned. Only the 'volume' key is used. Defaults to None.

        Returns:
            avail (dict): A VQIP of water available for abstraction
        """
        # Respond to abstraction check request
        avail = self.get_avail()

        if vqip:
            avail = self.v_change_vqip(avail, min(avail["volume"], vqip["volume"]))

        return avail

    def pull_set_abstraction(self, vqip):
        """Request set wrapper for get_avail where VQIP is specified.

        Args:
            vqip (dict): A VQIP of water to pull. Only the 'volume' key is used.

        Returns:
            avail (dict): A VQIP of water abstracted
        """
        # Respond to abstraction set request
        avail = self.get_avail()
        avail = self.v_change_vqip(avail, min(avail["volume"], vqip["volume"]))

        return avail

    def end_timestep_(self):
        """Reset unrouted water."""
        self.unrouted_water = self.empty_vqip()



# -*- coding: utf-8 -*-
"""Created on Mon Nov 15 14:20:36 2021.

@author: bdobson
Converted to totals on 2022-05-03
"""
from typing import Any, Dict

from wsimod.core import constants
from wsimod.nodes.nodes import Node
from wsimod.nodes.tanks import Tank


class WTW(Node):
    """A generic Water Treatment Works (WTW) node.

    This class is a generic water treatment works node. It is intended to be
    subclassed into freshwater and wastewater treatment works (FWTW and WWTW
    respectively).
    """

    def __init__(
        self,
        name,
        treatment_throughput_capacity=10,
        process_parameters={},
        liquor_multiplier={},
        percent_solids=0.0002,
    ):
        """Generic treatment processes that apply temperature a sensitive transform of
        pollutants into liquor and solids (behaviour depends on subclass). Push requests
        are stored in the current_input state variable, but treatment must be triggered
        with treat_current_input. This treated water is stored in the discharge_holder
        state variable, which will be sent different depending on FWTW/WWTW.

        Args:
            name (str): Node name
            treatment_throughput_capacity (float, optional): Amount of volume per
                timestep of water that can be treated. Defaults to 10.
            process_parameters (dict, optional): Dict of dicts for each pollutant.
                Top level key describes pollutant. Next level key describes the
                constant portion of the transform and the temperature sensitive
                exponent portion (see core.py/DecayObj for more detailed
                explanation). Defaults to {}.
            liquor_multiplier (dict, optional): Keys for each pollutant that
                describes how much influent becomes liquor. Defaults to {}.
            percent_solids (float, optional): Proportion of volume that becomes solids.
                All pollutants that do not become effluent or liquor become solids.
                Defaults to 0.0002.

        Functions intended to call in orchestration:
            None (use FWTW or WWTW subclass)

        Key assumptions:
            - Throughput can be modelled entirely with a set capacity.
            - Pollutant reduction for the entire treatment process can be modelled
                primarily with a single (temperature sensitive) transformation for
                each pollutant.
            - Liquor and solids are tracked and calculated with proportional
                multiplier parameters.

        Input data and parameter requirements:
            - `treatment_throughput_capacity`
                _Units_: cubic metres/timestep
            - `process_parameters` contains the constant (non temperature
                sensitive) and exponent (temperature sensitive) transformations
                applied to treated water for each pollutant.
                _Units_: -
            - `liquor_multiplier` and `percent_solids` describe the proportion of
                throughput that goes to liquor/solids.
        """
        # Set/Default parameters
        self.treatment_throughput_capacity = treatment_throughput_capacity
        if len(process_parameters) > 0:
            self.process_parameters = process_parameters
        else:
            self.process_parameters = {
                x: {"constant": 0.01, "exponent": 1.001}
                for x in constants.ADDITIVE_POLLUTANTS
            }
        if len(liquor_multiplier) > 0:
            self._liquor_multiplier = liquor_multiplier
        else:
            self._liquor_multiplier = {x: 0.7 for x in constants.ADDITIVE_POLLUTANTS}
            self._liquor_multiplier["volume"] = 0.03

        self._percent_solids = percent_solids

        # Update args
        super().__init__(name)

        self.process_parameters["volume"] = {"constant": self.calculate_volume()}

        # Update handlers
        self.push_set_handler["default"] = self.push_set_deny
        self.push_check_handler["default"] = self.push_check_deny

        # Initialise parameters
        self.current_input = self.empty_vqip()
        self.treated = self.empty_vqip()
        self.liquor = self.empty_vqip()
        self.solids = self.empty_vqip()

    def calculate_volume(self):
        """Calculate the volume proportion of treated water.

        Returns:
            (float): Volume of treated water
        """
        return 1 - self._percent_solids - self._liquor_multiplier["volume"]

    @property
    def percent_solids(self):
        return self._percent_solids

    @percent_solids.setter
    def percent_solids(self, value):
        self._percent_solids = value
        self.process_parameters["volume"]["constant"] = self.calculate_volume()

    @property
    def liquor_multiplier(self):
        return self._liquor_multiplier

    @liquor_multiplier.setter
    def liquor_multiplier(self, value):
        self._liquor_multiplier.update(value)
        self.process_parameters["volume"]["constant"] = self.calculate_volume()

    def apply_overrides(self, overrides=Dict[str, Any]):
        """Override parameters.

        Enables a user to override any of the following parameters:
        percent_solids, treatment_throughput_capacity, process_parameters (the
        entire dict does not need to be redefined, only changed values need to
        be included), liquor_multiplier (as with process_parameters).

        Args:
            overrides (Dict[str, Any]): Dict describing which parameters should
                be overridden (keys) and new values (values). Defaults to {}.
        """
        self.percent_solids = overrides.pop("percent_solids", self._percent_solids)
        self.liquor_multiplier = overrides.pop(
            "liquor_multiplier", self._liquor_multiplier
        )
        process_parameters = overrides.pop("process_parameters", {})
        for key, value in process_parameters.items():
            self.process_parameters[key].update(value)

        self.treatment_throughput_capacity = overrides.pop(
            "treatment_throughput_capacity", self.treatment_throughput_capacity
        )
        super().apply_overrides(overrides)

    def get_excess_throughput(self):
        """How much excess treatment capacity is there.

        Returns:
            (float): Amount of volume that can still be treated this timestep
        """
        return max(self.treatment_throughput_capacity - self.current_input["volume"], 0)

    def treat_current_input(self):
        """Run treatment processes this timestep, including temperature sensitive
        transforms, liquoring, solids."""
        # Treat current input
        influent = self.copy_vqip(self.current_input)

        # Calculate effluent, liquor and solids
        discharge_holder = self.empty_vqip()

        # Assume non-additive pollutants are unchanged in discharge and are
        # proportionately mixed in liquor
        for key in constants.NON_ADDITIVE_POLLUTANTS:
            discharge_holder[key] = influent[key]
            self.liquor[key] = (
                self.liquor[key] * self.liquor["volume"]
                + influent[key] * influent["volume"] * self.liquor_multiplier["volume"]
            ) / (
                self.liquor["volume"]
                + influent["volume"] * self.liquor_multiplier["volume"]
            )

        # TODO this should probably just be for process_parameters.keys() to avoid
        # having to declare non changing parameters
        # TODO should the liquoring be temperature sensitive too? As it is the solids
        # will take the brunt of the temperature variability which maybe isn't sensible
        for key in constants.ADDITIVE_POLLUTANTS + ["volume"]:
            if key != "volume":
                # Temperature sensitive transform
                temp_factor = self.process_parameters[key]["exponent"] ** (
                    constants.DECAY_REFERENCE_TEMPERATURE - influent["temperature"]
                )
            else:
                temp_factor = 1
            # Calculate discharge
            discharge_holder[key] = (
                influent[key] * self.process_parameters[key]["constant"] * temp_factor
            )
            # Calculate liquor
            self.liquor[key] = influent[key] * self.liquor_multiplier[key]

        # Calculate solids volume
        self.solids["volume"] = influent["volume"] * self.percent_solids

        # All remaining pollutants go to solids
        for key in constants.ADDITIVE_POLLUTANTS:
            self.solids[key] = influent[key] - discharge_holder[key] - self.liquor[key]

        # Blend with any existing discharge
        self.treated = self.sum_vqip(self.treated, discharge_holder)

        if self.treated["volume"] > self.current_input["volume"]:
            print("more treated than input")

    def end_timestep(self):
        """"""
        # Reset state variables
        self.current_input = self.empty_vqip()
        self.treated = self.empty_vqip()


class WWTW(WTW):
    """Wastewater Treatment Works (WWTW) node."""

    def __init__(
        self,
        stormwater_storage_capacity=10,
        stormwater_storage_area=1,
        stormwater_storage_elevation=10,
        **kwargs,
    ):
        """A wastewater treatment works wrapper for WTW. Contains a temporary stormwater
        storage tank. Liquor is combined with current_effluent and re- treated while
        solids leave the model.

        Args:
            stormwater_storage_capacity (float, optional): Capacity of stormwater tank.
                Defaults to 10.
            stormwater_storage_area (float, optional): Area of stormwater tank.
                Defaults to 1.
            stormwater_storage_elevation (float, optional): Datum of stormwater tank.
                Defaults to 10.

        Functions intended to call in orchestration:
            calculate_discharge

            make_discharge

        Key assumptions:
            - See `wtw.py/WTW` for treatment.
            - When `treatment_throughput_capacity` is exceeded, water is first sent
                to a stormwater storage tank before denying pushes. Leftover water
                in this tank aims to be treated in subsequent timesteps.
            - Can be pulled from to simulate active wastewater effluent use.

        Input data and parameter requirements:
            - See `wtw.py/WTW` for treatment.
            - Stormwater tank `capacity`, `area`, and `datum`.
                _Units_: cubic metres, squared metres, metres
        """
        # Set parameters
        self.stormwater_storage_capacity = stormwater_storage_capacity
        self.stormwater_storage_area = stormwater_storage_area
        self.stormwater_storage_elevation = stormwater_storage_elevation

        # Update args
        super().__init__(**kwargs)

        self.end_timestep = self.end_timestep_

        # Update handlers
        self.pull_set_handler["default"] = self.pull_set_reuse
        self.pull_check_handler["default"] = self.pull_check_reuse
        self.push_set_handler["Sewer"] = self.push_set_sewer
        self.push_check_handler["Sewer"] = self.push_check_sewer
        self.push_check_handler["default"] = self.push_check_sewer
        self.push_set_handler["default"] = self.push_set_sewer

        # Create tank
        self.stormwater_tank = Tank(
            capacity=self.stormwater_storage_capacity,
            area=self.stormwater_storage_area,
            datum=self.stormwater_storage_elevation,
        )

        # Initialise states
        self.liquor_ = self.empty_vqip()
        self.previous_input = self.empty_vqip()
        self.current_input = self.empty_vqip()  # TODO is this not done in WTW?

        # Mass balance
        self.mass_balance_out.append(lambda: self.solids)  # Assume these go to landfill
        self.mass_balance_ds.append(lambda: self.stormwater_tank.ds())
        self.mass_balance_ds.append(
            lambda: self.ds_vqip(self.liquor, self.liquor_)
        )  # Change in liquor

    def apply_overrides(self, overrides=Dict[str, Any]):
        """Apply overrides to the stormwater tank and WWTW.

        Enables a user to override any parameter of the stormwater tank, and
        then calls any overrides in WTW.

        Args:
            overrides (Dict[str, Any]): Dict describing which parameters should
                be overridden (keys) and new values (values). Defaults to {}.
        """
        self.stormwater_storage_capacity = overrides.pop(
            "stormwater_storage_capacity", self.stormwater_storage_capacity
        )
        self.stormwater_storage_area = overrides.pop(
            "stormwater_storage_area", self.stormwater_storage_area
        )
        self.stormwater_storage_elevation = overrides.pop(
            "stormwater_storage_elevation", self.stormwater_storage_elevation
        )
        self.stormwater_tank.area = self.stormwater_storage_area
        self.stormwater_tank.capacity = self.stormwater_storage_capacity
        self.stormwater_tank.datum = self.stormwater_storage_elevation
        super().apply_overrides(overrides)

    def calculate_discharge(self):
        """Clear stormwater tank if possible, and call treat_current_input."""
        # Run WWTW model

        # Try to clear stormwater
        # TODO (probably more tidy to use push_set_sewer? though maybe less
        # computationally efficient)
        excess = self.get_excess_throughput()
        if (self.stormwater_tank.get_avail()["volume"] > constants.FLOAT_ACCURACY) & (
            excess > constants.FLOAT_ACCURACY
        ):
            to_pull = min(excess, self.stormwater_tank.get_avail()["volume"])
            to_pull = self.v_change_vqip(self.stormwater_tank.storage, to_pull)
            cleared_stormwater = self.stormwater_tank.pull_storage(to_pull)
            self.current_input = self.sum_vqip(self.current_input, cleared_stormwater)

        # Run processes
        self.current_input = self.sum_vqip(self.current_input, self.liquor)
        self.treat_current_input()

    def make_discharge(self):
        """Discharge treated effluent."""
        reply = self.push_distributed(self.treated)
        self.treated = self.empty_vqip()
        if reply["volume"] > constants.FLOAT_ACCURACY:
            _ = self.stormwater_tank.push_storage(reply, force=True)
            print("WWTW couldnt push")

    def push_check_sewer(self, vqip=None):
        """Check throughput and stormwater tank capacity.

        Args:
            vqip (dict, optional): A VQIP that can be used to limit the volume in
                the return value (only volume key is used). Defaults to None.

        Returns:
            (dict): excess
        """
        # Get excess
        excess_throughput = self.get_excess_throughput()
        excess_tank = self.stormwater_tank.get_excess()
        # Combine tank and throughput
        vol = excess_tank["volume"] + excess_throughput
        # Update volume
        if vqip is None:
            vqip = self.empty_vqip()
        else:
            vol = min(vol, vqip["volume"])

        return self.v_change_vqip(vqip, vol)

    def push_set_sewer(self, vqip):
        """Receive water, first try to update current_input, and then stormwater tank.

        Args:
            vqip (dict): A VQIP amount to be treated and then stored

        Returns:
            (dict): A VQIP amount of water that was not treated
        """
        # Receive water from sewers
        vqip = self.copy_vqip(vqip)
        # Check if can directly be treated
        sent_direct = self.get_excess_throughput()

        sent_direct = min(sent_direct, vqip["volume"])

        sent_direct = self.v_change_vqip(vqip, sent_direct)

        self.current_input = self.sum_vqip(self.current_input, sent_direct)

        if sent_direct["volume"] == vqip["volume"]:
            # If all added to input, no problem
            return self.empty_vqip()

        # Next try temporary storage
        vqip = self.v_change_vqip(vqip, vqip["volume"] - sent_direct["volume"])

        vqip = self.stormwater_tank.push_storage(vqip)

        if vqip["volume"] < constants.FLOAT_ACCURACY:
            return self.empty_vqip()
        else:
            # TODO what to do here ???
            return vqip

    def pull_set_reuse(self, vqip):
        """Enables WWTW to receive pulls of the treated water (i.e., for wastewater
        reuse or satisfaction of environmental flows). Intended to be called in between
        calculate_discharge and make_discharge.

        Args:
            vqip (dict): A VQIP amount to be pulled (only 'volume' key is used)

        Returns:
            reply (dict): Amount of water that has been pulled
        """
        # Satisfy request with treated (volume)
        reply_vol = min(vqip["volume"], self.treated["volume"])
        # Update pollutants
        reply = self.v_change_vqip(self.treated, reply_vol)
        # Update treated
        self.treated = self.v_change_vqip(
            self.treated, self.treated["volume"] - reply_vol
        )
        return reply

    def pull_check_reuse(self, vqip=None):
        """Pull check available water. Simply returns the previous timestep's treated
        throughput. This is of course inaccurate (which may lead to slightly longer
        calulcations), but it is much more flexible. This hasn't been recently tested so
        it might be that returning treated would be fine (and more accurate!).

        Args:
            vqip (dict, optional): A VQIP that can be used to limit the volume in
                the return value (only volume key is used). Defaults to None.

        Returns:
            (dict): A VQIP amount of water available. Currently just the previous
                timestep's treated throughput
        """
        # Respond to request of water for reuse/MRF
        return self.copy_vqip(self.treated)

    def end_timestep_(self):
        """End timestep function to update state variables."""
        self.liquor_ = self.copy_vqip(self.liquor)
        self.previous_input = self.copy_vqip(self.current_input)
        self.current_input = self.empty_vqip()
        self.solids = self.empty_vqip()
        self.stormwater_tank.end_timestep()


class FWTW(WTW):
    """"""

    def __init__(
        self,
        service_reservoir_storage_capacity=10,
        service_reservoir_storage_area=1,
        service_reservoir_storage_elevation=10,
        service_reservoir_initial_storage=0,
        data_input_dict={},
        **kwargs,
    ):
        """A freshwater treatment works wrapper for WTW. Contains service reservoirs
        that treated water is released to and pulled from. Cannot allow deficit (thus
        any deficit is satisfied by water entering the model 'via other means'). Liquor
        and solids are sent to sewers.

        Args:
            service_reservoir_storage_capacity (float, optional): Capacity of service
                reservoirs. Defaults to 10.
            service_reservoir_storage_area (float, optional): Area of service
                reservoirs. Defaults to 1.
            service_reservoir_storage_elevation (float, optional): Datum of service
                reservoirs. Defaults to 10.
            service_reservoir_initial_storage (float or dict, optional): initial
                storage of service reservoirs (see nodes.py/Tank for details).
                Defaults to 0.
            data_input_dict (dict, optional): Dictionary of data inputs relevant for
                the node (though I don't think it is used). Defaults to {}.

        Functions intended to call in orchestration:
            treat_water

        Key assumptions:
            - See `wtw.py/WTW` for treatment.
            - Stores treated water in a service reservoir tank, with a single tank
                per `FWTW` node.
            - Aims to satisfy a throughput that would top up the service reservoirs
                until full.
            - Currently, will not allow a deficit, thus introducing water from
                'other measures' if pulls cannot fulfil demand. Behaviour under a
                deficit should be determined and validated before introducing.

        Input data and parameter requirements:
            - See `wtw.py/WTW` for treatment.
            - Service reservoir tank `capacity`, `area`, and `datum`.
                _Units_: cubic metres, squared metres, metres
        """
        # Default parameters
        self.service_reservoir_storage_capacity = service_reservoir_storage_capacity
        self.service_reservoir_storage_area = service_reservoir_storage_area
        self.service_reservoir_storage_elevation = service_reservoir_storage_elevation
        self.service_reservoir_initial_storage = service_reservoir_initial_storage
        # TODO don't think data_input_dict is used
        self.data_input_dict = data_input_dict

        # Update args
        super().__init__(**kwargs)
        self.end_timestep = self.end_timestep_

        # Update handlers
        self.pull_set_handler["default"] = self.pull_set_fwtw
        self.pull_check_handler["default"] = self.pull_check_fwtw

        self.push_set_handler["default"] = self.push_set_deny
        self.push_check_handler["default"] = self.push_check_deny

        # Initialise parameters
        self.total_deficit = self.empty_vqip()
        self.total_pulled = self.empty_vqip()
        self.previous_pulled = self.empty_vqip()
        self.unpushed_sludge = self.empty_vqip()

        # Create tanks
        self.service_reservoir_tank = Tank(
            capacity=self.service_reservoir_storage_capacity,
            area=self.service_reservoir_storage_area,
            datum=self.service_reservoir_storage_elevation,
            initial_storage=self.service_reservoir_initial_storage,
        )
        # self.service_reservoir_tank.storage['volume'] =
        # self.service_reservoir_inital_storage
        # self.service_reservoir_tank.storage_['volume'] =
        # self.service_reservoir_inital_storage

        # Mass balance
        self.mass_balance_in.append(lambda: self.total_deficit)
        self.mass_balance_ds.append(lambda: self.service_reservoir_tank.ds())
        self.mass_balance_out.append(lambda: self.unpushed_sludge)

    def apply_overrides(self, overrides=Dict[str, Any]):
        """Apply overrides to the service reservoir tank and FWTW.

        Enables a user to override any parameter of the service reservoir tank, and
        then calls any overrides in WTW.

        Args:
            overrides (Dict[str, Any]): Dict describing which parameters should
                be overridden (keys) and new values (values). Defaults to {}.
        """
        self.service_reservoir_storage_capacity = overrides.pop(
            "service_reservoir_storage_capacity",
            self.service_reservoir_storage_capacity,
        )
        self.service_reservoir_storage_area = overrides.pop(
            "service_reservoir_storage_area", self.service_reservoir_storage_area
        )
        self.service_reservoir_storage_elevation = overrides.pop(
            "service_reservoir_storage_elevation",
            self.service_reservoir_storage_elevation,
        )

        self.service_reservoir_tank.capacity = self.service_reservoir_storage_capacity
        self.service_reservoir_tank.area = self.service_reservoir_storage_area
        self.service_reservoir_tank.datum = self.service_reservoir_storage_elevation
        super().apply_overrides(overrides)

    def treat_water(self):
        """Pulls water, aiming to fill service reservoirs, calls WTW
        treat_current_input, avoids deficit, sends liquor and solids to sewers."""
        # Calculate how much water is needed
        target_throughput = self.service_reservoir_tank.get_excess()
        target_throughput = min(
            target_throughput["volume"], self.treatment_throughput_capacity
        )

        # Pull water
        throughput = self.pull_distributed({"volume": target_throughput})

        # Calculate deficit (assume is equal to difference between previous treated
        # throughput and current throughput)
        # TODO think about this a bit more
        deficit = max(target_throughput - throughput["volume"], 0)
        # deficit = max(self.previous_pulled['volume'] - throughput['volume'], 0)
        deficit = self.v_change_vqip(self.previous_pulled, deficit)

        # Introduce deficit
        self.current_input = self.sum_vqip(throughput, deficit)

        # Track deficit
        self.total_deficit = self.sum_vqip(self.total_deficit, deficit)

        if self.total_deficit["volume"] > constants.FLOAT_ACCURACY:
            print(
                "Service reservoirs not filled at {0} on {1}".format(self.name, self.t)
            )

        # Run treatment processes
        self.treat_current_input()

        # Discharge liquor and solids to sewers
        push_back = self.sum_vqip(self.liquor, self.solids)
        rejected = self.push_distributed(push_back, of_type="Sewer")
        self.unpushed_sludge = self.sum_vqip(self.unpushed_sludge, rejected)
        if rejected["volume"] > constants.FLOAT_ACCURACY:
            print("nowhere for sludge to go")

        # Send water to service reservoirs
        excess = self.service_reservoir_tank.push_storage(self.treated)
        _ = self.service_reservoir_tank.push_storage(excess, force=True)
        if excess["volume"] > 0:
            print("excess treated water")

    def pull_check_fwtw(self, vqip=None):
        """Pull checks query service reservoirs.

        Args:
            vqip (dict, optional): A VQIP that can be used to limit the volume in
                the return value (only volume key is used). Defaults to None.

        Returns:
            (dict): A VQIP of availability in service reservoirs
        """
        return self.service_reservoir_tank.get_avail(vqip)

    def pull_set_fwtw(self, vqip):
        """Pull treated water from service reservoirs.

        Args:
            vqip (dict): a VQIP amount to pull

        Returns:
            pulled (dict): A VQIP amount that was successfully pulled
        """
        # Pull
        pulled = self.service_reservoir_tank.pull_storage(vqip)
        # Update total_pulled this timestep
        self.total_pulled = self.sum_vqip(self.total_pulled, pulled)
        return pulled

    def end_timestep_(self):
        """Update state variables."""
        self.service_reservoir_tank.end_timestep()
        self.total_deficit = self.empty_vqip()
        self.previous_pulled = self.copy_vqip(self.total_pulled)
        self.total_pulled = self.empty_vqip()
        self.treated = self.empty_vqip()
        self.unpushed_sludge = self.empty_vqip()

    def reinit(self):
        """Call tank reinit."""
        self.service_reservoir_tank.reinit()



# -*- coding: utf-8 -*-
"""Created on Tue Oct 18 10:35:51 2022.

@author: Barney

"""

# import pytest
import unittest
from unittest import TestCase

from wsimod.arcs.arcs import Arc
from wsimod.core import constants
from wsimod.nodes.distribution import Distribution, UnlimitedDistribution
from wsimod.nodes.storage import Groundwater


class MyTestClass(TestCase):
    def setUp(self):
        """"""
        constants.set_simple_pollutants()

    def assertDictAlmostEqual(self, d1, d2, accuracy=19):
        """

        Args:
            d1:
            d2:
            accuracy:
        """
        for d in [d1, d2]:
            for key, item in d.items():
                d[key] = round(item, accuracy)
        self.assertDictEqual(d1, d2)

    def test_unlimited(self):
        distribution = UnlimitedDistribution(name="")

        d1 = {"volume": 1000, "phosphate": 0, "temperature": 0}
        reply = distribution.pull_set_unlimited(d1)

        self.assertDictAlmostEqual(d1, reply)
        self.assertDictAlmostEqual(d1, distribution.supplied)

        reply = distribution.pull_set_unlimited(d1)

        d1["volume"] = 2000
        self.assertDictAlmostEqual(d1, distribution.supplied)

    def test_leakage(self):
        distribution = Distribution(leakage=0.2, name="d1")

        udistribution = UnlimitedDistribution(name="u1")

        GW = Groundwater(name="g1", capacity=10)

        arc1 = Arc(name="a1", in_port=udistribution, out_port=distribution)

        arc2 = Arc(name="a2", in_port=distribution, out_port=GW)

        reply = distribution.pull_check({"volume": 5})
        v1 = 5
        self.assertEqual(v1, reply["volume"])

        reply = distribution.pull_set({"volume": 5})
        v2 = v1 / (1 - 0.2)
        self.assertEqual(v2, arc1.vqip_in["volume"])
        self.assertEqual(v2 * 0.2, arc2.vqip_in["volume"])

    def test_distribution_overrides(self):
        distribution = Distribution(name="", leakage=0.2)
        distribution.apply_overrides({"leakage": 0})
        self.assertEqual(distribution.leakage, 0)


if __name__ == "__main__":
    unittest.main()



from wsimod.extensions import register_node_patch


@register_node_patch("my_fwtw", "pull_distributed")
def custom_pull_distributed(self, vqip, *args, **kwargs):
    """A custom `pull_distributed` function.

    Call `pull_distributed` with the tag "FWTW".
    """
    return self._patched_pull_distributed(vqip, tag="FWTW")
